

import warnings


import numpy as np
import matplotlib.pyplot as plt
import datastock as ds


from ._DataCollection_class0_Base import DataCollection0
from . import _DataCollection_interactivity as _interactivity
from . import _DataCollection_comp


__all__ = ['DataCollection1']    # , 'TimeTraceCollection']


_DKEYS = {
    'control': {'val': False, 'action': 'generic'},
    'ctrl': {'val': False, 'action': 'generic'},
    'shift': {'val': False, 'action': 'generic'},
    'alt': {'val': False, 'action': 'generic'},
    'left': {'val': False, 'action': 'move'},
    'right': {'val': False, 'action': 'move'},
    'up': {'val': False, 'action': 'move'},
    'down': {'val': False, 'action': 'move'},
}
_INCREMENTS = [1, 10]


# #################################################################
# #################################################################
#               Main class
# #################################################################


class DataCollection1(DataCollection0):
    """ Handles matplotlib interactivity """

    _LPAXES = ['ax', 'type']
    _dinteractivity = dict.fromkeys(['curax_panzoom'])

    # ----------------------
    #   Add objects
    # ----------------------

    def add_mobile(
        self,
        key=None,
        handle=None,
        ref=None,
        data=None,
        dtype=None,
        bstr=None,
        visible=None,
        ax=None,
        **kwdargs,
    ):

        # ----------
        # check ref

        if isinstance(ref, str):
            ref = (ref,)
        if isinstance(ref, list):
            ref = tuple(ref)

        if ref is None or not all([rr in self._dref.keys() for rr in ref]):
            msg = (
                "Arg ref must be a tuple of existing ref keys!\n"
                f"\t- Provided: {ref}"
            )
            raise Exception(msg)

        # ----------
        # check dtype

        dtype = ds._generic_check._check_var(
            dtype,
            'dtype',
            types=str,
            allowed=['xdata', 'ydata', 'data', 'alpha', 'txt']
        )

        # ----------
        # check data

        if isinstance(data, str):
            data = (data,)
        if isinstance(data, list):
            data = tuple(data)
        if data is None:
            data = ['index' for rr in ref]

        c0 = (
            len(ref) == len(data)
            and all([rr == 'index' or rr in self._ddata.keys() for rr in data])
        )
        if not c0:
            msg = (
                "Arg data must be a tuple of existing data keys!\n"
                "It should hqve the same length as ref!\n"
                f"\t- Provided ref: {ref}\n"
                f"\t- Provided data: {data}"
            )
            raise Exception(msg)

        super().add_obj(
            which='mobile',
            key=key,
            handle=handle,
            group=None,
            ref=ref,
            data=data,
            dtype=dtype,
            visible=visible,
            bstr=bstr,
            ax=ax,
            func=None,
            **kwdargs,
        )

    def add_axes(
        self,
        key=None,
        handle=None,
        type=None,
        refx=None,
        refy=None,
        datax=None,
        datay=None,
        invertx=None,
        inverty=None,
        **kwdargs,
    ):

        # ----------------
        # check refx, refy

        if isinstance(refx, str):
            refx = [refx]
        if isinstance(refy, str):
            refy = [refy]

        c0 = (
            isinstance(refx, list)
            and all([rr in self._dref.keys() for rr in refx])
        )
        if refx is not None and not c0:
            msg = "Arg refx must be a list of valid ref keys!"
            raise Exception(msg)

        c0 = (
            isinstance(refy, list)
            and all([rr in self._dref.keys() for rr in refy])
        )
        if refy is not None and not c0:
            msg = "Arg refy must be a list of valid ref keys!"
            raise Exception(msg)

        # data
        if datax is None and refx is not None:
            datax = ['index' for rr in refx]
        if datay is None and refy is not None:
            datay = ['index' for rr in refy]

        super().add_obj(
            which='axes',
            key=key,
            handle=handle,
            type=type,
            groupx=None,
            groupy=None,
            refx=refx,
            refy=refy,
            datax=datax,
            datay=datay,
            invertx=invertx,
            inverty=inverty,
            bck=None,
            mobile=None,
            canvas=None,
            **kwdargs,
        )

        # add canvas if not already stored
        if 'canvas' not in self._dobj.keys():
            self.add_canvas(handle=handle.figure.canvas)
        else:
            lisin = [
                k0 for k0, v0 in self._dobj['canvas'].items()
                if v0['handle'] == handle.figure.canvas
            ]
            if len(lisin) == 0:
                self.add_canvas(handle=handle.figure.canvas)

    def add_canvas(self, key=None, handle=None):
        """ Add canvas and interactivity obj """
        interactive = (
            hasattr(handle, 'toolbar')
            and handle.toolbar is not None
        )
        self.add_obj(
            which='canvas',
            key=key,
            handle=handle,
            interactive=interactive,
        )

    # ------------------
    # Properties
    # ------------------

    @property
    def dax(self):
        return self.dobj.get('axes', {})

    @property
    def dinteractivity(self):
        return self._dinteractivity

    # ------------------
    # Debug mode
    # ------------------

    def set_debug(self, debug=None):
        """ Set debug mode to True / False """
        debug = ds._generic_check._check_var(
            debug,
            'debug',
            default=False,
            types=bool,
        )
        self.debug = debug

    def show_debug(self):
        """ Display information relevant for live debugging """
        print('\n\n')
        return self.get_summary(
            show_which=['ref', 'group', 'interactivity'],
        )

    # ------------------
    # Setup interactivity
    # ------------------

    def setup_interactivity(
        self,
        kinter=None,
        dgroup=None,
        dkeys=None,
        dinc=None,
        cur_ax=None,
        debug=None,
    ):
        """

        dgroup = {
            'group0': {
                'ref': ['ref0', 'ref1', ...],
                'nmax': 3,
                'colors': ['r', 'g', 'b'],
            }
        }

        """

        # ----------
        # Check dgroup

        c0 = (
            isinstance(dgroup, dict)
            and all([
                isinstance(k0, str)
                and isinstance(v0, dict)
                and isinstance(v0.get('ref'), list)
                and isinstance(v0.get('data'), list)
                and len(v0['ref']) == len(v0['data'])
                and all([ss in self._dref.keys() for ss in v0['ref']])
                for k0, v0 in dgroup.items()
            ])
        )
        if not c0:
            msg = "Arg dgroup must be a dict of the form:\n"
            raise Exception(msg)

        ic = 0
        for k0, v0 in dgroup.items():
            if v0.get('nmax') is None:
                dgroup[k0]['nmax'] = 0
            dgroup[k0]['nmaxcur'] = 0
            dgroup[k0]['indcur'] = 0

        # ----------
        # Check increment dict

        if dinc is None:
            dinc = {k0: _INCREMENTS for k0 in self._dref.keys()}
        elif isinstance(dinc, list) and len(dinc) == 2:
            dinc = {k0: dinc for k0 in self._dref.keys()}
        elif isinstance(dinc, dict):
            c0 = all([
                ss in self._dref.keys()
                and isinstance(vv, list)
                and len(vv) == 2
                for ss, vv in dinc.items()
            ])
            if not c0:
                msg = (
                    "Arg dinc must be a dict of type {ref0: [inc0, inc1]}\n"
                    f"\t- Provided: {dinc}"
                )
                raise Exception(msg)
            for k0 in self._dref.keys():
                if k0 not in dinc.keys():
                    dinc[k0] = _INCREMENTS
        else:
            msg = (
                "Arg dinc must be a dict of type {ref0: [inc0, inc1]}\n"
                f"\t- Provided: {dinc}"
            )
            raise Exception(msg)

        # ----------------------------
        # make sure all refs are known

        drefgroup = dict.fromkeys(self._dref.keys())
        for k0, v0 in self._dref.items():
            lg = [k1 for k1, v1 in dgroup.items() if k0 in v1['ref']]
            if len(lg) > 1:
                msg = f"Ref {k0} has no/several groups!\n\t- found: {lg}"
                raise Exception(msg)
            elif len(lg) == 0:
                lg = [None]
            drefgroup[k0] = lg[0]

            #  add indices
            if lg[0] is not None:
                self.add_indices_per_ref(
                    indices=np.zeros((dgroup[lg[0]]['nmax'],), dtype=int),
                    ref=k0,
                    distribute=False,
                )

        self.add_param(which='ref', param='group', value=drefgroup)
        self.add_param(which='ref', param='inc', value=dinc)

        # --------------------------------------
        # update dax with groupx, groupy and inc

        daxgroupx = dict.fromkeys(self._dobj['axes'].keys())
        daxgroupy = dict.fromkeys(self._dobj['axes'].keys())
        dinc = dict.fromkeys(self._dobj['axes'].keys())
        for k0, v0 in self._dobj['axes'].items():
            if v0['refx'] is None:
                daxgroupx[k0] = None
            else:
                daxgroupx[k0] = [
                    self._dref[k1]['group'] for k1 in v0['refx']
                ]
            if v0['refy'] is None:
                daxgroupy[k0] = None
            else:
                daxgroupy[k0] = [
                    self._dref[k1]['group'] for k1 in v0['refy']
                ]

            # increment
            dinc[k0] = {
                'left': -1, 'right': 1,
                'down': -1, 'up': 1,
            }

        self.set_param(which='axes', param='groupx', value=daxgroupx)
        self.set_param(which='axes', param='groupy', value=daxgroupy)
        self.add_param(which='axes', param='inc', value=dinc)

        # group, ref, nmax

        # cumsum0 = np.r_[0, np.cumsum(nmax[1, :])]
        # arefind = np.zeros((np.sum(nmax[1, :]),), dtype=int)

        # --------------------------
        # update mobile with groups

        for k0, v0 in self._dobj['mobile'].items():
            self._dobj['mobile'][k0]['group'] = tuple([
                self._dref[rr]['group'] for rr in v0['ref']
            ])
            self._dobj['mobile'][k0]['func'] = _interactivity.get_fupdate(
                handle=v0['handle'],
                dtype=v0['dtype'],
                norm=None,
                bstr=v0.get('bstr'),
            )

        # --------------------
        # axes mobile, refs and canvas

        daxcan = dict.fromkeys(self._dobj['axes'].keys())
        for k0, v0 in self._dobj['axes'].items():

            # Update mobile
            self._dobj['axes'][k0]['mobile'] = [
                k1 for k1, v1 in self._dobj['mobile'].items()
                if v1['ax'] == k0
            ]

            # ref
            if v0['refx'] is not None:
                for ii, rr in enumerate(v0['refx']):
                    if v0['datax'][ii] is None:
                        self._dobj['axes'][k0]['datax'][ii] = 'index'

            # canvas
            lcan = [
                k1 for k1, v1 in self._dobj['canvas'].items()
                if v1['handle'] == v0['handle'].figure.canvas
            ]
            assert len(lcan) == 1
            self._dobj['axes'][k0]['canvas'] = lcan[0]

        # -------
        # dgroup

        # update with axes
        for k0, v0 in dgroup.items():
            lkax = [
                k1 for k1, v1 in self._dobj['axes'].items()
                if (v1['groupx'] is not None and k0 in v1['groupx'])
                or (v1['groupy'] is not None and k0 in v1['groupy'])
            ]
            dgroup[k0]['axes'] = lkax

        for k0, v0 in dgroup.items():
            self.add_obj(
                which='group',
                key=k0,
                **v0,
            )

        # ---------
        # dkeys

        if dkeys is None:
            dkeys = _DKEYS

        # add key for switching groups
        dkeys.update({
            v0.get('key', f'f{ii+1}'): {
                'group': k0,
                'val': False,
                'action': 'group',
            }
            for ii, (k0, v0) in enumerate(self._dobj['group'].items())
        })

        # add keys for switching indices within groups
        nMax = np.max([v0['nmax'] for v0 in dgroup.values()])
        dkeys.update({
            str(ii): {'ind': ii, 'val': False, 'action': 'indices'}
            for ii in range(0, nMax)
        })

        # implement dict
        for k0, v0 in dkeys.items():
            self.add_obj(
                which='key',
                key=k0,
                **v0,
            )

        lact = set([v0['action'] for v0 in dkeys.values()])
        self.__dkeys_r = {
            k0: [k1 for k1 in dkeys.keys() if dkeys[k1]['action'] == k0]
            for k0 in lact
        }

        # ---------
        # dinter

        dinter = {
            'cur_ax': cur_ax,
            'cur_groupx': None,
            'cur_groupy': None,
            'cur_refx': None,
            'cur_refy': None,
            'cur_datax': None,
            'cur_datay': None,
            'follow': True,
        }

        self.add_obj(
            which='interactivity',
            key='inter0',
            **dinter,
        )
        self.kinter = 'inter0'

        _interactivity._set_dbck(
            lax=self._dobj['axes'].keys(),
            daxes=self._dobj['axes'],
            dcanvas=self._dobj['canvas'],
            dmobile=self._dobj['mobile'],
        )

        # -----------------------------------
        # set current axe / group / ref / data...

        if cur_ax is None:
            cur_ax = [
                k0 for k0, v0 in self._dobj['axes'].items()
                if v0['groupx'] is not None and v0['groupy'] is not None
            ]
            if len(cur_ax) == 0:
                cur_ax = list(self._dobj['axes'].keys())[0]
            else:
                cur_ax = cur_ax[0]

        self._get_current_grouprefdata_from_kax(kax=cur_ax)
        self.set_debug(debug)

    # ----------------------------
    # Ensure connectivity possible
    # ----------------------------

    def _warn_ifnotInteractive(self):
        warn = False

        c0 = (
            len(self._dobj.get('axes', {})) > 0
            and len(self._dobj.get('canvas', {})) > 0
        )
        if c0:
            dcout = {
                k0: v0['handle'].__class__.__name__
                for k0, v0 in self._dobj['canvas'].items()
                if not v0['interactive']
            }
            if len(dcout) > 0:
                lstr = '\n'.join(
                    [f'\t- {k0}: {v0}' for k0, v0 in dcout.items()]
                )
                msg = (
                    "Non-interactive backends identified (prefer Qt5Agg):\n"
                    f"\t- backend : {plt.get_backend()}\n"
                    f"\t- canvas  :\n{lstr}"
                )
                warn = True
        else:
            msg = ("No available axes / canvas for interactivity")
            warn = True

        # raise warning
        if warn:
            warnings.warn(msg)

        return warn

    # ----------------------
    # Connect / disconnect
    # ----------------------

    def connect(self):
        if self._warn_ifnotInteractive():
            return
        for k0, v0 in self._dobj['canvas'].items():
            keyp = v0['handle'].mpl_connect('key_press_event', self.onkeypress)
            keyr = v0['handle'].mpl_connect('key_release_event', self.onkeypress)
            butp = v0['handle'].mpl_connect('button_press_event', self.mouseclic)
            # res = v0['handle'].mpl_connect('resize_event', self.resize)
            butr = v0['handle'].mpl_connect('button_release_event', self.mouserelease)
            close = v0['handle'].mpl_connect('close_event', self.on_close)
            # if not plt.get_backend() == "agg":
            # v0['handle'].manager.toolbar.release = self.mouserelease

            self._dobj['canvas'][k0]['cid'] = {
                'keyp': keyp,
                'keyr': keyr,
                'butp': butp,
                # 'res': res,
                'butr': butr,
                'close': close,
            }

    def disconnect(self):
        if self._warn_ifnotInteractive():
            return
        for k0, v0 in self._dobj['canvas'].items():
            for k1, v1 in v0['cid'].items():
                v0['handle'].mpl_disconnect(v1)
            v0['handle'].manager.toolbar.release = lambda event: None

    # ------------------------------------
    # Interactivity handling - preliminary
    # ------------------------------------

    def _get_current_grouprefdata_from_kax(self, kax=None):

        # Get current group and ref
        groupx = self._dobj['axes'][kax]['groupx']
        groupy = self._dobj['axes'][kax]['groupy']
        refx = self._dobj['axes'][kax]['refx']
        refy = self._dobj['axes'][kax]['refy']

        # Get kinter
        kinter = list(self._dobj['interactivity'].keys())[0]

        # Get current groups
        cur_groupx = self._dobj['interactivity'][kinter]['cur_groupx']
        cur_groupy = self._dobj['interactivity'][kinter]['cur_groupy']

        # determine whether cur_groupx shall be updated
        if groupx is not None:
            if cur_groupx in groupx:
                pass
            elif groupx is not None:
                cur_groupx = groupx[0]
        if groupy is not None:
            if cur_groupy in groupy:
                pass
            elif groupy is not None:
                cur_groupy = groupy[0]

        # # get current refs
        cur_refx = self._dobj['interactivity'][kinter]['cur_refx']
        cur_refy = self._dobj['interactivity'][kinter]['cur_refy']
        if groupx is not None:
            if cur_refx in self._dobj['group'][cur_groupx]['ref']:
                pass
            elif cur_groupx is not None:
                cur_refx = self._dobj['group'][cur_groupx]['ref'][0]
        if groupy is not None:
            if cur_refy in self._dobj['group'][cur_groupy]['ref']:
                pass
            elif cur_groupy is not None:
                cur_refy = self._dobj['group'][cur_groupy]['ref'][0]

        # data
        cur_datax = self._dobj['interactivity'][kinter]['cur_datax']
        if self._dobj['axes'][kax]['refx'] is not None:
            ix = self._dobj['axes'][kax]['refx'].index(cur_refx)
            cur_datax = self._dobj['axes'][kax]['datax'][ix]

        cur_datay = self._dobj['interactivity'][kinter]['cur_datay']
        if self._dobj['axes'][kax]['refy'] is not None:
            iy = self._dobj['axes'][kax]['refy'].index(cur_refy)
            cur_datay = self._dobj['axes'][kax]['datay'][iy]

        # Update interactivity dict
        self.kinter = kinter
        self._dobj['interactivity'][kinter].update({
            'cur_ax': kax,
            'ax_panzoom': kax,
            'cur_groupx': cur_groupx,
            'cur_groupy': cur_groupy,
            'cur_refx': cur_refx,
            'cur_refy': cur_refy,
            'cur_datax': cur_datax,
            'cur_datay': cur_datay,
        })

    def _getset_current_axref(self, event):
        # Check click is relevant
        c0 = event.inaxes is not None and event.button == 1
        if not c0:
            raise Exception("clic not in axes")

        # get current ax key
        lkax = [
            k0 for k0, v0 in self._dobj['axes'].items()
            if v0['handle'] == event.inaxes
        ]
        kax = ds._generic_check._check_var(
            None, 'kax',
            types=str,
            allowed=lkax,
        )
        ax = self._dobj['axes'][kax]['handle']

        # Check axes is relevant and toolbar not active
        lc = [
            all([
                'fix' not in v0.keys()
                for v0 in self._dobj['axes'].values()
            ]),
            all([
                not v0['handle'].manager.toolbar.mode
                for v0 in self._dobj['canvas'].values()
            ]),
        ]
        if not all(lc):
            raise Exception("Not usable axes!")

        self._get_current_grouprefdata_from_kax(kax=kax)

    # -----------------------------
    # Interactivity: generic update
    # -----------------------------

    def update_interactivity(self):
        """ Called at each event """

        cur_groupx = self._dobj['interactivity'][self.kinter]['cur_groupx']
        cur_groupy = self._dobj['interactivity'][self.kinter]['cur_groupy']
        cur_refx = self._dobj['interactivity'][self.kinter]['cur_refx']
        cur_refy = self._dobj['interactivity'][self.kinter]['cur_refy']
        cur_datax = self._dobj['interactivity'][self.kinter]['cur_datax']
        cur_datay = self._dobj['interactivity'][self.kinter]['cur_datay']

        # Propagate indices through refs
        if cur_refx is not None:
            lref = self._dobj['group'][cur_groupx]['ref']
            ldata = self._dobj['group'][cur_groupx]['data']
            self.propagate_indices_per_ref(
                ref=cur_refx,
                lref=[rr for rr in lref if rr != cur_refx],
                ldata=[cur_datax] + [dd for dd in ldata if dd != cur_datax],
                param=None,
            )

        if cur_refy is not None:
            lref = self._dobj['group'][cur_groupy]['ref']
            ldata = self._dobj['group'][cur_groupy]['data']
            self.propagate_indices_per_ref(
                ref=cur_refy,
                lref=[rr for rr in lref if rr != cur_refy],
                ldata=[cur_datay] + [dd for dd in ldata if dd != cur_datay],
                param=None,
            )

        # get list of mobiles to update and set visible
        lmobiles = []
        if cur_groupx is not None:
            lmobiles += [
                k0 for k0, v0 in self._dobj['mobile'].items()
                if any([
                    rr in v0['ref']
                    for rr in self._dobj['group'][cur_groupx]['ref']
                ])
            ]

        if cur_groupy is not None:
            lmobiles += [
                k0 for k0, v0 in self._dobj['mobile'].items()
                if any([
                    rr in v0['ref']
                    for rr in self._dobj['group'][cur_groupy]['ref']
                ])
            ]

        self._update_mobiles(lmobiles=lmobiles)     # 0.2 s

        if self.debug:
            self.show_debug()

    def _update_mobiles(self, lmobiles=None):

        # Set visibility of mobile objects - TBF/TBC
        for k0 in lmobiles:
            vis = all([
                self._dobj['mobile'][k0]['ind']
                < self._dobj['group'][gg]['nmaxcur']
                for gg in self._dobj['mobile'][k0]['group']
            ])
            self._dobj['mobile'][k0]['visible'] = vis

        # get list of axes to update
        lax = [
            k0 for k0, v0 in self._dobj['axes'].items()
            if any([self._dobj['mobile'][k1]['ax'] == k0 for k1 in lmobiles])
        ]

        # ---- Restore backgrounds ---- 1 ms
        for aa in lax:
            self._dobj['canvas'][
                self._dobj['axes'][aa]['canvas']
            ]['handle'].restore_region(
                self._dobj['axes'][aa]['bck'],
            )

        # ---- update data of group objects ----  0.15 s
        for k0 in lmobiles:
            _interactivity._update_mobile(
                dmobile=self._dobj['mobile'],
                dref=self._dref,
                ddata=self._ddata,
                k0=k0,
            )

        # --- Redraw all objects (due to background restore) --- 25 ms
        for k0, v0 in self._dobj['mobile'].items():
            v0['handle'].set_visible(v0['visible'])
            self._dobj['axes'][v0['ax']]['handle'].draw_artist(v0['handle'])

        # ---- blit axes ------ 5 ms
        for aa in lax:
            self._dobj['canvas'][
                self._dobj['axes'][aa]['canvas']
            ]['handle'].blit(self._dobj['axes'][aa]['handle'].bbox)

    def resize(self, event):
        _interactivity._set_dbck(
            lax=self._dobj['axes'].keys(),
            daxes=self._dobj['axes'],
            dcanvas=self._dobj['canvas'],
            dmobile=self._dobj['mobile'],
        )

    # ----------------------
    # Interactivity: mouse
    # ----------------------

    def mouseclic(self, event):

        # Check click is relevant
        c0 = event.button == 1
        if not c0:
            return

        # get / set cuurent interactive usabel axes and ref
        try:
            self._getset_current_axref(event)
        except Exception as err:
            if str(err) == 'clic not in axes':
                return
            raise err
            # warnings.warn(str(err))
            # return

        kinter = self.kinter
        kax = self._dobj['interactivity'][kinter]['cur_ax']

        refx = self._dobj['axes'][kax]['refx']
        refy = self._dobj['axes'][kax]['refy']
        if refx is None and refy is None:
            return

        cur_groupx = self._dobj['interactivity'][kinter]['cur_groupx']
        cur_groupy = self._dobj['interactivity'][kinter]['cur_groupy']
        cur_refx = self._dobj['interactivity'][kinter]['cur_refx']
        cur_refy = self._dobj['interactivity'][kinter]['cur_refy']
        cur_datax = self._dobj['interactivity'][kinter]['cur_datax']
        cur_datay = self._dobj['interactivity'][kinter]['cur_datay']

        shift = self._dobj['key']['shift']['val']
        ctrl = any([
            self._dobj['key'][ss]['val'] for ss in ['control', 'ctrl']
        ])

        # Update number of indices (for visibility)
        for gg in [cur_groupx, cur_groupy]:
            if gg is not None:
                out = _interactivity._update_indices_nb(
                    group=gg,
                    dgroup=self._dobj['group'],
                    ctrl=ctrl,
                    shift=shift,
                )

        # update indcur x/y vs shift / ctrl ?  TBF
        pass

        # Check refx/refy vs datax/datay
        if cur_refx is not None and cur_refy is not None:
            c0 = (
                'index' in [cur_datax, cur_datay]
                or ((cur_refx == cur_refy) == (cur_datax == cur_datay))
            )
            if not c0:
                msg = (
                    "Invalid ref / data pairs:\n"
                    f"\t- cur_refx, cur_refy: {cur_refx}, {cur_refy}\n"
                    f"\t- cdx, cdy: {cdx}, {cdy}"
                )
                raise Exception(msg)

        # update ref indices
        if None not in [cur_refx, cur_refy] and cur_refx == cur_refy:

            raise NotImplementedError()
            dist = (cdx - event.xdata)**2 + (cdy - event.ydata)**2
            lind = [
                np.nanargmin(dist, axis=ii) for ii in range(datax.ndim)
            ]

        else:

            c0x = (
                cur_refx is not None
                and self._dobj['axes'][kax]['refx'] is not None
                and cur_refx in self._dobj['axes'][kax]['refx']
            )
            if c0x:
                monot = None
                if cur_datax == 'index':
                    cdx = 'index'
                else:
                    monot = self._ddata[cur_datax]['monot'] == (True,)
                    cdx = self._ddata[cur_datax]['data']
                ix = _DataCollection_comp._get_index_from_data(
                    data=cdx,
                    data_pick=np.r_[event.xdata],
                    monot=monot,
                )[0]

            c0y = (
                cur_refy is not None
                and self._dobj['axes'][kax]['refy'] is not None
                and cur_refy in self._dobj['axes'][kax]['refy']
            )
            if c0y:
                monot = None
                if cur_datay == 'index':
                    cdy = 'index'
                else:
                    monot = self._ddata[cur_datay]['monot'] == (True,)
                    cdy = self._ddata[cur_datay]['data']
                iy = _DataCollection_comp._get_index_from_data(
                    data=cdy,
                    data_pick=np.r_[event.ydata],
                    monot=monot,
                )[0]

        # Update ref indices
        if c0x:
            cur_ix = self._dobj['group'][cur_groupx]['indcur']
            follow = (
                cur_ix == self._dobj['group'][cur_groupx]['nmaxcur'] - 1
                and self._dobj['interactivity'][kinter]['follow']
            )
            if follow:
                self._dref[cur_refx]['indices'][cur_ix:] = ix
            else:
                self._dref[cur_refx]['indices'][cur_ix] = ix

        # Update ref indices
        if c0y:
            cur_iy = self._dobj['group'][cur_groupy]['indcur']
            follow = (
                cur_iy == self._dobj['group'][cur_groupy]['nmaxcur'] - 1
                and self._dobj['interactivity'][kinter]['follow']
            )
            if follow:
                self._dref[cur_refy]['indices'][cur_iy:] = iy
            else:
                self._dref[cur_refy]['indices'][cur_iy] = iy

        self.update_interactivity()

    def mouserelease(self, event):
        """ Mouse release: nothing except if resize ongoing (redraw bck) """

        c0 = event.inaxes is not None and event.button == 1
        if not c0:
            return

        can = [
            k0 for k0, v0 in self._dobj['canvas'].items()
            if v0['handle'] == event.inaxes.figure.canvas
        ][0]
        mode = self._dobj['canvas'][can]['handle'].manager.toolbar.mode.lower()
        c0 = 'pan' in mode
        c1 = 'zoom' in mode

        if c0 or c1:
            kax = self._dobj['interactivity']['curax_panzoom']
            if kax is None:
                msg = (
                    "Make sure you release the mouse button on an axes!"
                    "\n Otherwise background plot cannot be properly updated!"
                )
                raise Exception(msg)
            ax = self._dobj['axes'][kax]['handle']
            lax = ax.get_shared_x_axes().get_siblings(ax)
            lax += ax.get_shared_y_axes().get_siblings(ax)
            lax = list(set(lax))
            _interactivity._set_dbck(
                lax=lax,
                daxes=self._dobj['axes'],
                dcanvas=self._dobj['canvas'],
                dmobile=self._dobj['mobile'],
            )

    # ----------------------
    # Interactivity: keys
    # ----------------------

    def onkeypress(self, event):
        """ Event handler in case of key press / release """

        # -----------------------
        # Check event is relevant 1

        # decompose key combinations
        lkey = event.key.split('+')

        # get current inter, axes, canvas
        kinter = self.kinter
        kax = self._dobj['interactivity'][kinter]['cur_ax']
        kcan = [
            k0 for k0, v0 in self._dobj['canvas'].items()
            if k0 == self._dobj['axes'][kax]['canvas']
        ][0]
        can = self._dobj['canvas'][kcan]['handle']

        # check relevance
        c0 = can.manager.toolbar.mode != ''
        c1 = len(lkey) not in [1, 2]
        c2 = [ss not in self._dobj['key'].keys() for ss in lkey]

        if c0 or c1 or any(c2):
            return

        # -----------------------
        # Check event is relevant 2

        # get list of current keys for each action type
        lgen = [kk for kk in self.__dkeys_r['generic'] if kk in lkey]
        lmov = [kk for kk in self.__dkeys_r['move'] if kk in lkey]
        lgrp = [kk for kk in self.__dkeys_r['group'] if kk in lkey]
        lind = [kk for kk in self.__dkeys_r['indices'] if kk in lkey]

        # if no relevant key pressed => return
        ngen, nmov, ngrp, nind = len(lgen), len(lmov), len(lgrp), len(lind)
        ln = np.r_[ngen, nmov, ngrp, nind]
        if np.any(ln > 1) or np.sum(ln) > 2:
            return
        if np.sum(ln) == 2 and (ngrp == 1 or nind == 1):
            return

        # only keep relevant keys
        genk = None if ngen == 0 else lgen[0]
        movk = None if nmov == 0 else lmov[0]
        grpk = None if ngrp == 0 else lgrp[0]
        indk = None if nind == 0 else lind[0]

        # ------------------------
        # Event = change key value

        # change key values if relevant
        if event.name == 'key_release_event':
            if event.key == genk:
                self._dobj['key'][genk]['val'] = False
            return

        if genk is not None and event.key == genk:
            self._dobj['key'][genk]['val'] = True
            return

        # ----------------------------
        # Event = change current group

        if grpk is not None:
            # group
            group = self._dobj['key'][event.key]['group']
            cx = any([
                v0['groupx'] is not None and group in v0['groupx']
                for v0 in self._dobj['axes'].values()
            ])
            if cx:
                self._dobj['interactivity'][self.kinter]['cur_groupx'] = group
            cy = any([
                v0['groupy'] is not None and group in v0['groupy']
                for v0 in self._dobj['axes'].values()
            ])
            if cy:
                self._dobj['interactivity'][self.kinter]['cur_groupy'] = group

            # axes
            cur_ax = self._dobj['interactivity'][self.kinter]['cur_ax']
            if cur_ax not in self._dobj['group'][group]['axes']:
                self._dobj['interactivity'][self.kinter]['cur_ax'] = (
                    self._dobj['group'][group]['axes'][0]
                )

            # ref
            if cx:
                cur_refx = self._dobj['interactivity'][self.kinter]['cur_refx']
                if self._dref[cur_refx]['group'] != group:
                    cur_refx = self._dobj['group'][group]['ref'][0]
                self._dobj['interactivity'][self.kinter]['cur_refx'] = cur_refx

            if cy:
                cur_refy = self._dobj['interactivity'][self.kinter]['cur_refy']
                if self._dref[cur_refy]['group'] != group:
                    cur_refy = self._dobj['group'][group]['ref'][0]
                self._dobj['interactivity'][self.kinter]['cur_refy'] = cur_refy

            # data
            if c0:
                self._dobj['interactivity'][self.kinter]['cur_datax'] = 'index'
            if c0:
                self._dobj['interactivity'][self.kinter]['cur_datay'] = 'index'

            msg = f"Current group set to {group}"
            print(msg)
            return

        # ----------------------------
        # Event = change current index

        if indk is not None:
            groupx = self._dobj['interactivity'][self.kinter]['cur_groupx']
            groupy = self._dobj['interactivity'][self.kinter]['cur_groupy']

            # groupx
            if groupx is not None:
                imax = self._dobj['group'][groupx]['nmaxcur']
                ii = int(event.key)
                if ii > imax:
                    msg = "Set to current max index for '{groupx}': {imax}"
                    print(msg)
                ii = min(ii, imax)
                self._dobj['group'][groupx]['indcur'] = ii

            # groupy
            if groupy is not None:
                imax = self._dobj['group'][groupy]['nmaxcur']
                ii = int(event.key)
                if ii > imax:
                    msg = "Set to current max index for '{groupy}': {imax}"
                    print(msg)
                ii = min(ii, imax)
                self._dobj['group'][groupy]['indcur'] = ii

            msg = f"Current indices set to {ii}"
            print(msg)
            return

        # ----------------------------
        # Event = move current index

        if movk is not None:

            if movk in ['left', 'right']:
                group = self._dobj['interactivity'][self.kinter]['cur_groupx']
                ref = self._dobj['interactivity'][self.kinter]['cur_refx']
                incsign = 1.
                if self._dobj['axes'][kax].get('invertx', False):
                    incsign = -1
            elif movk in ['up', 'down']:
                group = self._dobj['interactivity'][self.kinter]['cur_groupy']
                ref = self._dobj['interactivity'][self.kinter]['cur_refy']
                incsign = 1.
                if self._dobj['axes'][kax].get('inverty', False):
                    incsign = -1

            if group is None:
                return

            # dmovkeys for inversions and steps ?

            shift = self._dobj['key']['shift']['val']
            ctrl = any([
                self._dobj['key'][ss]['val'] for ss in ['control', 'ctrl']
            ])
            alt = self._dobj['key']['alt']['val']

            # Check max number of occurences not reached if shift
            c0 = (
                shift
                and (
                    self._dobj['group'][group]['indcur']
                    == self._dobj['group'][group]['nmax'] - 1
                )
            )
            if c0:
                msg = "Max nb. of plots reached ({0}) for group {1}"
                msg = msg.format(self._dobj['group'][group]['nmax'], group)
                print(msg)
                return

            # update nb of visible indices
            out = _interactivity._update_indices_nb(
                group=group,
                dgroup=self._dobj['group'],
                ctrl=ctrl,
                shift=shift,
            )
            if out is False:
                return

            # get increment from key
            cax = self._dobj['interactivity'][self.kinter]['cur_ax']
            inc = incsign * (
                self._dref[ref]['inc'][int(alt)]
                * self._dobj['axes'][cax]['inc'][movk]
            )

            # update ref indices
            icur = self._dobj['group'][group]['indcur']
            ix = (
                (self._dref[ref]['indices'][icur] + inc)
                % self._dref[ref]['size']
            )

            # Update ref indices
            follow = (
                icur == self._dobj['group'][group]['nmaxcur'] - 1
                and self._dobj['interactivity'][kinter]['follow']
            )
            if follow:
                self._dref[ref]['indices'][icur:] = ix
            else:
                self._dref[ref]['indices'][icur] = ix

            # global update of interactivity
            self.update_interactivity()

    # -------------------
    # Close all
    # -------------------

    def on_close(self, event):
        kcan = [
            k0 for k0, v0 in self._dobj['canvas'].items()
            if v0['handle'] == event.canvas
        ]
        if len(kcan) > 1:
            raise Exception('Several matching canvas')
        elif len(kcan) == 1:

            if len(self._dobj['canvas']) == 1:
                self.close_all()

            else:
                lax = [
                    k1 for k1, v1 in self._dobj['axes'].items()
                    if v1['canvas'] == kcan[0]
                ]
                lmob = [
                    k1 for k1, v1 in self._dobj['mobile'].items()
                    if v1['ax'] in lax
                ]
                for k1 in lax:
                    del self._dobj['axes'][k1]
                for k1 in lmob:
                    del self._dobj['mobile'][k1]
                del self._dobj['canvas'][kcan[0]]

    def close_all(self):

        # close figures
        if 'axes' in self._dobj.keys():
            lfig = set([
                v0['handle'].figure for v0 in self._dobj['axes'].values()
            ])
            for ff in lfig:
                plt.close(ff)

        # delete obj dict
        lk = ['interactivity', 'mobile', 'key', 'canvas', 'group', 'axes']
        for kk in lk:
            if kk in self._dobj.keys():
                del self._dobj[kk]

        # remove interactivity-specific param in dref
        lp = list(set(self.get_lparam(which='ref')).intersection(
            ['indices', 'group', 'inc']
        ))
        self.remove_param(which='ref', param=lp)




import os
import warnings


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt


_HERE = os.path.dirname(__file__)
_TOFU_PATH = os.path.dirname(os.path.dirname(os.path.dirname(_HERE)))

# Import current tofu version
cwd = os.getcwd()
os.chdir(_TOFU_PATH)
import tofu as tf
os.chdir(_HERE)


# -----------------------------------------------------------------------------
#                   Default parameters
# -----------------------------------------------------------------------------


# _HERE = os.path.abspath(os.path.dirname(__file__))
_PFE = os.path.join(_HERE, 'WEST_Geometry_Rev_Nov2016_V2.xlsx')

# Toroidal width of a sector
_DTHETA = 2*np.pi / 18.

# 4 corners of a Casing Cover PJ from Creoview
_PJ_PTS = np.array([[313.728, -988.400, 2255.087],
                    [476.477, -988.400, 2226.390],
                    [313.728, -988.400, 1938.128],
                    [368.071, -988.400, 1928.547]])
_PJ_THETA = np.arctan2(-_PJ_PTS[:, 2], _PJ_PTS[:, 0])
_PJ_DTHETA = np.abs(0.5*((_PJ_THETA[1]-_PJ_THETA[0])
                         + (_PJ_THETA[3]-_PJ_THETA[2])))
_PJ_POS0 = (0.5*((_PJ_THETA[1]+_PJ_THETA[0])/2 + (_PJ_THETA[3]+_PJ_THETA[2])/2)
            - _DTHETA)
_PJ_POS = (_PJ_POS0
           + np.r_[0, 1, 3, 4, 6, 7, 9, 10, 12, 13, 15, 16]*_DTHETA)
_CASING_DTHETA = [_DTHETA - _PJ_DTHETA, 2*_DTHETA - _PJ_DTHETA]
_CASING_POS = [_PJ_POS0 + _DTHETA/2 + np.arange(6)*3*_DTHETA,
               _PJ_POS0 + 2*_DTHETA + np.arange(6)*3*_DTHETA]
_CASING_DTHETA = np.tile(_CASING_DTHETA, 6)
_CASING_POS = np.array(_CASING_POS).T.ravel()

# Port plug holes
_PPLUG_PTS = 1.e-3*np.array([
    [-895.804, 420.123, 3063.270],
    [-895.804, -420.123, 3063.270],
    [-205.919, -420.123, 3184.915],
    [-205.919, 420.123, 3184.915],
])
_PPLUG_THETA = np.arctan2(-_PPLUG_PTS[:, 2], _PPLUG_PTS[:, 0])
_PPLUG_DTHETA = _PPLUG_THETA.max() - _PPLUG_THETA.min()
_PPLUG_POS = 0.5*(_PPLUG_THETA.max() + _PPLUG_THETA.min())
_PPLUG_POS = _PPLUG_POS + _DTHETA*np.r_[0, np.cumsum(np.tile([1, 2], 6))[:-1]]
_PPLUG_FULL_EXTENT0 = _DTHETA - _PPLUG_DTHETA
_PPLUG_FULL_EXTENT1 = 2.*_DTHETA - _PPLUG_DTHETA
_PPLUG_FULL_POS0 = _PPLUG_POS[0] + _DTHETA*(0.5 + 3*np.r_[0, 1, 2, 3, 4, 5])
_PPLUG_FULL_POS1 = _PPLUG_POS[1] + _DTHETA*(1 + 3*np.r_[0, 1, 2, 3, 4, 5])


# Default
_EXP = 'WEST'
_DNAMES = {
    'Baffle': {
        'name': None,
    },
    'LPA': {
        'name': None,
    },
    'VDE': {
        'name': None,
    },
    # Vacuum vessels
    'Inner VV': {
        'name': 'InnerV0',
        'class': 'Ves',
        'save': True,
    },
    'Outer VV': {
        'name': 'OuterV0',
        'class': 'Ves',
        'save': True,
    },
    # Thermal shields
    'IVPP HFS': {
        'name': 'ThermalShieldHFSV0',
        'class': 'PFC',
        'thick': 0.008,
        'thickpos': False,
        'save': True,
    },
    'IVPP LFS': {
        'name': 'ThermalShieldLFSV0',
        'class': 'PFC',
        'thick': 0.008,
        'thickpos': True,
        'save': False,      # Will be replaced by several parts
    },
    # Lower divertor
    'LDiv PFUs': {
        'name': None,
    },
    'LDiv CasingCover': {
        'name': 'CasingCoverLDivV0',
        'class': 'PFC',
        'save': True,
    },
    'Ldiv Casing': {
        'name': 'CasingLDivV0',
        'class': 'PFC',
        'extent': _CASING_DTHETA,
        'pos': _CASING_POS,
        'save': True,
    },
    'Ldiv Casing PJ': {
        'name': 'CasingPJLDivV0',
        'class': 'PFC',
        'extent': _PJ_DTHETA,
        'pos': _PJ_POS,
        'save': True,
    },
    'Ldiv PFU Plate': {
        'name': 'CasingPFUPlateLDivV0',
        'class': 'PFC',
        'save': True,
    },
    # Upper divertor
    'UDiv PFUs': {
        'name': None,
    },
    'UDiv CasingCover': {
        'name': 'CasingCoverUDivV0',
        'class': 'PFC',
        'save': True,
    },
    'Udiv Casing': {
        'name': 'CasingUDivV0',
        'class': 'PFC',
        'extent': _CASING_DTHETA,
        'pos': _CASING_POS,
        'save': True,
    },
    'Udiv Casing PJ': {
        'name': 'CasingPJUDivV0',
        'class': 'PFC',
        'extent': _PJ_DTHETA,
        'pos': _PJ_POS,
        'save': True,
    },
    'Udiv PFU Plate': {
        'name': 'CasingPFUPlateUDivV0',
        'class': 'PFC',
        'save': True,
    },
}


# #############################################################################
# #############################################################################
#           Define PEI
# #############################################################################


def get_PEI_HFS():

    pts = np.array([
        [-1681.880, 555.372, 784.274],
        [-1670.349, 494.641, 778.897],
        [-1661.932, 444.534, 774.972],
        [-1652.797, 383.409, 770.712],
        [-1642.850, 303.089, 766.074],
        [-1635.270, 223.416, 762.539],
        [-1630.704, 155.868, 760.410],
        [-1627.334, 76.573, 758.838],
        [-1626.250, 2.920, 758.333],
        [-1627.273, -74.958, 758.810],
        [-1630.443, -151.695, 760.288],
        [-1635.199, -222.607, 762.506],
        [-1641.223, -288.671, 765.315],
        [-1649.738, -361.293, 769.285],
        [-1662.010, -445.008, 775.008],
        [-1670.139, -493.533, 778.799],
        [-1681.880, -555.372, 784.274],
    ])

    poly = 1.e-3*np.array([np.hypot(pts[:, 0], pts[:, 2]), pts[:, 1]])
    return poly


def get_PEI_LFS(poly):

    poly = poly[:, ~np.any(np.isnan(poly), axis=0)]
    _, indsZ = np.unique(poly[1, :], return_index=True)
    poly = poly[:, indsZ]
    poly_pplug = np.array([np.hypot(_PPLUG_PTS[:, 0], -_PPLUG_PTS[:, 2]),
                           _PPLUG_PTS[:, 1]])
    _, indspp = np.unique(poly_pplug[1, :], return_index=True)
    poly_pplug = poly_pplug[:, indspp]
    poly_pplug[0, :] = np.interp(poly_pplug[1, :], poly[1, :], poly[0, :])
    ind_insert = np.searchsorted(poly[1, :], poly_pplug[1, :])
    poly_full = np.insert(poly, ind_insert, poly_pplug, axis=1)
    ilow = np.nonzero(poly_full[1, :] == poly_pplug[1, 0])[0][0]
    iup = np.nonzero(poly_full[1, :] == poly_pplug[1, 1])[0][0]
    poly_low = poly_full[:, :ilow+1]
    poly_up = poly_full[:, iup:]

    dpeilfs = {
        'full_slim': {
            'name': 'ThermalShieldLFSSlimV0',
            'class': 'PFC',
            'poly': poly_full,
            'pos': _PPLUG_FULL_POS0,
            'extent': _PPLUG_FULL_EXTENT0,
            'thick': 0.008,
            'thickpos': True,
            'save': True,
        },
        'full_wide': {
            'name': 'ThermalShieldLFSWideV0',
            'class': 'PFC',
            'poly': poly_full,
            'pos': _PPLUG_FULL_POS1,
            'extent': _PPLUG_FULL_EXTENT1,
            'thick': 0.008,
            'thickpos': True,
            'save': True,
        },
        'pplug_low': {
            'name': 'ThermalShieldLFSLowV0',
            'class': 'PFC',
            'poly': poly_low,
            'pos': _PPLUG_POS,
            'extent': _PPLUG_DTHETA,
            'thick': 0.008,
            'thickpos': True,
            'save': True,
        },
        'pplug_up': {
            'name': 'ThermalShieldLFSUpV0',
            'class': 'PFC',
            'poly': poly_up,
            'pos': _PPLUG_POS,
            'extent': _PPLUG_DTHETA,
            'thick': 0.008,
            'thickpos': True,
            'save': True,
        },
    }
    return dpeilfs


# #############################################################################
# #############################################################################
#           Extract and plot geometry
# #############################################################################


def _add_thickness(poly, thick, pos=None, nn=None):
    dv = poly[:, 1:] - poly[:, :-1]
    vout = np.array([dv[1, :], -dv[0, :]])
    if pos and np.nanmean(vout[0, :]) < 0:
        vout = -vout
    elif (~pos) and np.nanmean(vout[0, :]) > 0:
        vout = -vout
    vout = np.concatenate((vout[:, 0:1], vout), axis=1)
    vout = vout / np.sqrt(np.sum(vout**2, axis=0))[None, :]
    poly = np.concatenate(
        (poly, poly[:, ::-1] + thick*vout[:, ::-1]),
        axis=1
    )
    return poly


def _create_poly_obj(dnames=None, out=None, nn=None, ss=None):

    # Create / check poly
    if dnames[nn].get('poly') is None:
        poly = np.array([out[ss]['R (m)'], out[ss]['Z (m)']])
        if nn == 'IVPP HFS':
            poly = get_PEI_HFS()
        dnames[nn]['poly'] = poly
    indok = ~np.any(np.isnan(dnames[nn]['poly']), axis=0)
    dnames[nn]['poly'] = dnames[nn]['poly'][:, indok]
    if 'thick' in dnames[nn].keys():
        dnames[nn]['poly'] = _add_thickness(
            dnames[nn]['poly'],
            dnames[nn]['thick'],
            pos=dnames[nn]['thickpos'],
            nn=nn,
        )

    # Create obj
    if 'class' in dnames[nn].keys():
        try:
            obj = getattr(tf.geom, dnames[nn]['class'])(
                Name=dnames[nn]['name'],
                Poly=dnames[nn]['poly'],
                Exp=_EXP,
                pos=dnames[nn].get('pos', None),
                extent=dnames[nn].get('extent', None),
            )
        except Exception as err:
            obj = None
            msg = (str(err)
                   + "\ntofu object {} failed".format(nn))
            warnings.warn(str(err))
        dnames[nn]['obj'] = obj
    return


def get_all(
    pfe=None,
    dnames=None,
    ax=None,
    plot=None,
    save=None,
    return_ax=None,
):

    # --------------
    #   Plot
    if pfe is None:
        pfe = _PFE
    if dnames is None:
        dnames = {kk: dict(vv) for kk, vv in _DNAMES.items()}
    if plot is None:
        plot = True
    if save is None:
        save = False
    if return_ax is None:
        return_ax = False

    # --------------
    #   Extract
    out = pd.read_excel(pfe, sheet_name='Main', header=[0, 1])

    # Add split LFS thermal shield
    dnames.update(get_PEI_LFS(
        np.array([
            out['IVPP LFS\n//IVPP_LFS//']['R (m)'],
            out['IVPP LFS\n//IVPP_LFS//']['Z (m)'],
        ])
    ))

    # From out
    ls = list(out.columns.levels[0])
    for ss in ls:
        nn = ss.split('\n')[0]
        if nn not in dnames.keys():
            continue
        _create_poly_obj(dnames=dnames, out=out, nn=nn, ss=ss)

    # From dnames
    for nn in dnames.keys():
        if 'obj' in dnames[nn].keys():
            continue
        _create_poly_obj(dnames=dnames, out=out, nn=nn, ss=None)

    # --------------
    #   Plot
    lobj = [nn for nn in dnames.keys() if dnames[nn].get('obj') is not None]
    if ax is None:
        if len(lobj) > 0:
            lax = None
            for oo in lobj:
                lax = dnames[oo]['obj'].plot(
                    lax=lax,
                    proj='all',
                    element='P',
                    indices=False,
                    dLeg=None,
                )
            ax = lax[0]
        else:
            fig = plt.figure(figsize=(10, 10))
            ax = fig.add_axes([0.1, 0.1, 0.8, 0.8], aspect='equal')

    for nn in dnames.keys():
        ll, = ax.plot(dnames[nn]['poly'][0, :], dnames[nn]['poly'][1, :],
                      label=nn)

    ax.legend(loc='upper left', bbox_to_anchor=(1., 1.))

    # --------------
    #   save
    if save is True:
        for nn in dnames.keys():
            c0 = (dnames[nn].get('save') is True
                  and dnames[nn].get('obj') is not None)
            if c0:
                dnames[nn]['obj'].save_to_txt(path=_HERE)

    if return_ax is True:
        return dnames, ax
    else:
        return dnames



# -*- coding: utf-8 -*-


# Built-in
# import copy


# Common
import numpy as np
import datastock as ds


# tofu
from ._class08_Diagnostic import Diagnostic as Previous
from . import _class9_compute as _compute
from . import _class9_plot as _plot


__all__ = ['GeometryMatrix']


# #############################################################################
# #############################################################################
#                           Diagnostic
# #############################################################################


class GeometryMatrix(Previous):

    _show_in_summary = 'all'

    _dshow = dict(Previous._dshow)
    _dshow.update({
        'geom matrix': [
            'bsplines',
            'diagnostic',
            'camera',
            'shape',
            'data',
            'method',
            'res',
        ],
    })

    # -----------------
    # geometry matrix
    # ------------------

    def add_geometry_matrix(
        self,
        key=None,
        key_bsplines=None,
        key_diag=None,
        key_cam=None,
        # sampling
        res=None,
        mode=None,
        method=None,
        crop=None,
        dvos=None,
        # common ref
        ref_com=None,
        ref_vector_strategy=None,
        # options
        brightness=None,
        # output
        verb=None,
        store=None,
    ):

        return _compute.compute(
            coll=self,
            key=key,
            key_bsplines=key_bsplines,
            key_diag=key_diag,
            key_cam=key_cam,
            # sampling
            res=res,
            mode=mode,
            method=method,
            crop=crop,
            dvos=dvos,
            # common ref
            ref_com=ref_com,
            ref_vector_strategy=ref_vector_strategy,
            # options
            brightness=brightness,
            # output
            verb=verb,
            store=store,
        )

    # -----------------
    # get concatenated geometry matrix
    # ------------------

    def get_geometry_matrix_concatenated(
        self,
        key=None,
        key_cam=None,
    ):
        """ Assemble the geometry matrix """

        return _compute._concatenate(
            coll=self,
            key=key,
            key_cam=key_cam,
        )

    # -----------------
    # plotting
    # ------------------

    def plot_geometry_matrix(
        self,
        key=None,
        indbf=None,
        indchan=None,
        # options
        plot_mesh=None,
        plot_config=None,
        # parameters
        vmin=None,
        vmax=None,
        res=None,
        cmap=None,
        # figure
        dax=None,
        dmargin=None,
        fs=None,
        dcolorbar=None,
        dleg=None,
    ):
        return _plot.plot_geometry_matrix(
            coll=self,
            key=key,
            indbf=indbf,
            indchan=indchan,
            # options
            plot_mesh=plot_mesh,
            plot_config=plot_config,
            # parameters
            vmin=vmin,
            vmax=vmax,
            res=res,
            cmap=cmap,
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            dcolorbar=dcolorbar,
            dleg=dleg,
        )


# -*- coding: utf-8 -*-


# Built-in


# Common
import numpy as np
import datastock as ds


# specific
from . import _generic_check
from . import _generic_plot


# ###############################################################
# ###############################################################
#                           plot main
# ###############################################################


def _plot_rays(
    coll=None,
    key=None,
    proj=None,
    res=None,
    mode=None,
    concatenate=None,
    # config
    plot_config=None,
    # figure
    dax=None,
    dmargin=None,
    fs=None,
    wintit=None,
    # interactivity
    color_dict=None,
    nlos=None,
    dinc=None,
    connect=None,
):

    # ------------
    # check inputs

    (
        key,
        ref,
        mode,
        concatenate,
        proj,
        color_dict,
        nlos,
        connect,
    ) = _plot_rays_check(
        coll=coll,
        key=key,
        mode=mode,
        concatenate=concatenate,
        # figure
        proj=proj,
        # interactivity
        color_dict=color_dict,
        nlos=nlos,
        connect=connect,
    )

    # -------------------------
    # prepare los interactivity

    rays_x, rays_y, rays_z = coll.sample_rays(
        key=key,
        res=res,
        mode=mode,
        concatenate=concatenate,
    )
    if concatenate is False and rays_x.ndim > 2:
        shape = (rays_x.shape[0], -1)
        rays_x = rays_x.reshape(shape)
        rays_y = rays_y.reshape(shape)
        rays_z = rays_z.reshape(shape)

    rays_r = np.hypot(rays_x, rays_y)

    # -----------------
    # prepare figure

    if dax is None:

        dax = _generic_plot.get_dax_diag(
            proj=proj,
            dmargin=dmargin,
            fs=fs,
            wintit=wintit,
            tit=key,
        )

    dax = _generic_check._check_dax(dax=dax, main=proj[0])

    # -----------------
    # plot static parts

    # cross
    kax = 'cross'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        if concatenate is True:
            ax.plot(
                rays_r,
                rays_z,
                c='k',
                lw=1.,
                ls='-',
            )
        else:
            ax.plot(
                rays_r,
                rays_z,
                lw=1.,
                ls='-',
            )

    # hor
    kax = 'hor'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        if concatenate is True:
            ax.plot(
                rays_x,
                rays_y,
                c='k',
                lw=1.,
                ls='-',
            )
        else:
            ax.plot(
                rays_x,
                rays_y,
                lw=1.,
                ls='-',
            )

    # 3d
    kax = '3d'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        if concatenate is True:
            ax.plot(
                rays_x,
                rays_y,
                rays_z,
                c='k',
                lw=1.,
                ls='-',
            )
        else:
            for ii in range(rays_x.shape[1]):
                ax.plot(
                    rays_x[:, ii],
                    rays_y[:, ii],
                    rays_z[:, ii],
                    lw=1.,
                    ls='-',
                )

    # -------
    # config

    if plot_config.__class__.__name__ == 'Config':

        kax = 'cross'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']
            plot_config.plot(lax=ax, proj=kax)

        kax = 'hor'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']
            plot_config.plot(lax=ax, proj=kax)

    return dax


# ###############################################################
# ###############################################################
#                           plot check
# ###############################################################


def _plot_rays_check(
    coll=None,
    key=None,
    mode=None,
    concatenate=None,
    # figure
    proj=None,
    # interactivity
    color_dict=None,
    nlos=None,
    connect=None,
):

    # -------
    # key

    lok = list(coll.dobj.get('rays', {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )

    ref = coll.dobj['rays'][key]['ref']

    # ------------
    # concatenate

    concatenate = ds._generic_check._check_var(
        concatenate, 'concatenate',
        types=bool,
        default=True,
    )

    # -----
    # mode

    lok = ['rel']
    if concatenate is True:
        lok += ['abs']
    mode = ds._generic_check._check_var(
        mode, 'mode',
        types=str,
        default='rel',
        allowed=lok,
    )

    # -----
    # proj

    proj = _generic_plot._proj(
        proj=proj,
        pall=['cross', 'hor', '3d'],
    )

    # -------
    # color_dict

    if color_dict is None:
        lc = ['r', 'g', 'b', 'm', 'c', 'y']
        color_dict = {
            'x': lc,
            'y': lc,
        }

    # -------
    # nlos

    nlos = ds._generic_check._check_var(
        nlos, 'nlos',
        types=int,
        default=5,
    )

    # -------
    # connect

    connect = ds._generic_check._check_var(
        connect, 'connect',
        types=bool,
        default=True,
    )

    return (
        key,
        ref,
        mode,
        concatenate,
        proj,
        color_dict,
        nlos,
        connect,
    )





import numpy as np
import scipy.spatial as scpsp
from matplotlib.path import Path
import datastock as ds


# ########################################################
# ########################################################
#              DEFAULTS
# ########################################################


_DANGLES = {
    'angle0': {
        'nb': 101,
        'lim': np.r_[-1, 1] * 90,
        'units': 'deg',
    },
    'angle1': {
        'nb': 101,
        'lim': np.r_[-1, 1] * 90,
        'units': 'deg',
    },
}


# ########################################################
# ########################################################
#              Main
# ########################################################


def main(
    coll=None,
    key=None,
    # from rays
    key_rays=None,
    segment=None,
    # from user-defined
    cent=None,
    nin=None,
    e0=None,
    e1=None,
    # angles
    angle0=None,
    angle1=None,
    config=None,
    strict=None,
    # optional naming
    key_angle0=None,
    key_angle1=None,
    ref_angle0=None,
    ref_angle1=None,
    units_angles=None,
):

    # -------------
    # check inputs
    # -------------

    key, cent, nin, e0, e1, dangles, strict = _check(
        coll=coll,
        key=key,
        # from rays
        key_rays=key_rays,
        segment=segment,
        # user-defined
        cent=cent,
        nin=nin,
        e0=e0,
        e1=e1,
        # angles
        angle0=angle0,
        angle1=angle1,
        strict=strict,
        # optional naming
        key_angle0=key_angle0,
        key_angle1=key_angle1,
        ref_angle0=ref_angle0,
        ref_angle1=ref_angle1,
        units_angles=units_angles,
    )

    # -------------
    # prepare
    # -------------

    # -----------
    # angles ref

    for k0, v0 in dangles.items():
        if v0['ref'] not in coll.dref.keys():
            coll.add_ref(v0['ref'], size=v0['data'].size)

    # ------------
    # angles data

    for k0, v0 in dangles.items():
        if v0['key'] not in coll.ddata.keys():
            coll.add_data(**v0)

    # -------------
    # unit vectors
    # -------------

    # angles, full shape
    angle0f = dangles['angle0']['data'][:, None]
    if dangles['angle0']['units'] == 'deg':
        angle0f = angle0f * np.pi/180.

    angle1f = dangles['angle1']['data'][None, :]
    if dangles['angle1']['units'] == 'deg':
        angle1f = angle1f * np.pi/180.

    # unit vectors
    vx = (
        np.cos(angle1f)
        * (np.cos(angle0f) * (nin[0]) + np.sin(angle0f) * e0[0])
        + np.sin(angle1f) * e1[0]
    )
    vy = (
        np.cos(angle1f)
        * (np.cos(angle0f) * (nin[1]) + np.sin(angle0f) * e0[1])
        + np.sin(angle1f) * e1[1]
    )
    vz = (
        np.cos(angle1f)
        * (np.cos(angle0f) * (nin[2]) + np.sin(angle0f) * e0[2])
        + np.sin(angle1f) * e1[2]
    )

    # -------------
    # compute
    # -------------

    coll.add_rays(
        key,
        start_x=np.full(vx.shape, cent[0]),
        start_y=np.full(vx.shape, cent[1]),
        start_z=np.full(vx.shape, cent[2]),
        # vectors
        vect_x=vx,
        vect_y=vy,
        vect_z=vz,
        # ref
        ref=(dangles['angle0']['ref'], dangles['angle1']['ref']),
        # specific to ptcam
        angles=(dangles['angle0']['key'], dangles['angle1']['key']),
        nin=nin,
        e0=e0,
        e1=e1,
        # config
        config=config,
        strict=strict,
    )

    return


# ########################################################
# ########################################################
#              Check
# ########################################################


def _check(
    coll=None,
    key=None,
    # from rays
    key_rays=None,
    segment=None,
    # from user-defined
    cent=None,
    nin=None,
    e0=None,
    e1=None,
    angle0=None,
    angle1=None,
    strict=None,
    # optional naming
    key_angle0=None,
    key_angle1=None,
    ref_angle0=None,
    ref_angle1=None,
    units_angles=None,
):

    # --------------
    # key
    # --------------

    wrays = coll._which_rays
    lout = list(coll.dobj.get(wrays, {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        default='ptcam',
        excluded=lout,
    )

    # ---------------------
    # rays vs user-defined
    # ---------------------

    lc = [
        key_rays is not None,
        cent is not None,
    ]
    if np.sum(lc) != 1:
        msg = (
            "Please provide either key_rays xor cent!\n"
            f"\t- key_rays: {key_rays}\n"
            f"\t- cent: {cent}\n"
        )
        raise Exception(msg)

    # --------------
    # key_rays
    # --------------

    if key_rays is not None:

        if isinstance(key_rays, str):
            key_rays = [key_rays]

        # key_rays
        lok = list(coll.dobj.get(wrays, {}).keys())
        key_rays = ds._generic_check._check_var_iter(
            key_rays, 'key_rays',
            types=list,
            types_iter=str,
            allowed=lok,
        )

        # segment
        nseg = [coll.dobj[wrays][kk]['shape'][0] for kk in key_rays]
        segment = ds._generic_check._check_var(
            segment, 'segment',
            types=int,
            default=0,
            allowed=list(range(np.min(nseg))),
        )

        # derive cent
        cent, nini = _rays_intersection(
            coll,
            key_rays=key_rays,
            segment=segment,
        )

        # nin
        if nin is None:
            nin = nini

    # --------------
    # cent
    # --------------

    else:

        cent = ds._generic_check._check_flat1darray(
            cent, 'cent',
            dtype=float,
            size=3,
        )

    # --------------
    # unit vectors
    # --------------

    nin, e0, e1 = ds._generic_check._check_vectbasis(
        e0=nin,
        e1=e0,
        e2=e1,
        dim=3,
        direct=False,
    )

    # --------------
    # angle0, angle1 angles
    # --------------

    dangles = {
        'angle0': _check_angles(
            coll=coll,
            key=key,
            ang=angle0,
            ang_name='angle0',
            ang_key=key_angle0,
            ang_ref=ref_angle0,
        ),
        'angle1': _check_angles(
            coll=coll,
            key=key,
            ang=angle1,
            ang_name='angle1',
            ang_key=key_angle1,
            ang_ref=ref_angle1,
        ),
    }

    # --------------
    # options
    # --------------

    # strict
    strict = ds._generic_check._check_var(
        strict, 'strict',
        types=bool,
        default=False,
    )

    return key, cent, nin, e0, e1, dangles, strict


def _check_angles(
    coll=None,
    key=None,
    ang=None,
    ang_name=None,
    ang_key=None,
    ang_ref=None,
    ang_units=None,
):

    # --------------
    # existing key
    # --------------

    if isinstance(ang, str):
        lok = [
            k0 for k0, v0 in coll.ddata.items()
            if v0['monot'] == (True,)
        ]
        ang = ds._generic_check._check_var(
            ang, ang_name,
            types=str,
            allowed=lok,
        )

        ang_key = ang
        ang_data = coll.ddata[ang_key]['data']
        ang_ref = coll.ddata[ang_key]['ref'][0]
        ang_units = coll.ddata[ang_key]['units']

    # --------------
    # int, float, array
    # --------------

    else:

        if ang is None:
            ang = _DANGLES[ang_name]['nb']

        if isinstance(ang, (float, int)):
            ang = np.linspace(
                _DANGLES[ang_name]['lim'][0],
                _DANGLES[ang_name]['lim'][1],
                int(ang),
            )
            ang_units = _DANGLES[ang_name]['units']

        # --------------
        # flat unique array

        ang_data = ds._generic_check._check_flat1darray(
            ang, ang_name,
            dtype=float,
            unique=True,
        )

        # -------
        # keys

        lout = list(coll.ddata.keys())
        ang_key = ds._generic_check._check_var(
            ang_key, ang_name,
            types=str,
            default=f"{key}_{ang_name}",
            excluded=lout,
        )

        # -------
        # ref

        lout = list(coll.dref.keys())
        ang_ref = ds._generic_check._check_var(
            ang_ref, f"{ang_name}",
            types=str,
            default=f"{key}_n{ang_name}",
            excluded=lout,
        )

        # -------
        # units

        ang_units = ds._generic_check._check_var(
            ang_units, "units_angles",
            types=str,
            default="deg",
            allowed=['deg', 'rad'],
        )

    return {
        'key': ang_key,
        'data': ang_data,
        'ref': ang_ref,
        'units': ang_units,
    }


# ##################################################
# ##################################################
#          Get rays intersection
# ##################################################


def _rays_intersection(
    coll,
    key_rays=None,
    segment=None,
):

    # ----------------
    # prepare
    # ----------------

    wrays = coll._which_rays
    nrays = len(key_rays)
    nrays_tot = np.array([
        np.prod(coll.dobj[wrays][kray]['shape'][1:]) for kray in key_rays
    ])

    cx = np.full((nrays,), np.nan)
    cy = np.full((nrays,), np.nan)
    cz = np.full((nrays,), np.nan)

    ux = np.full((nrays,), np.nan)
    uy = np.full((nrays,), np.nan)
    uz = np.full((nrays,), np.nan)

    # ----------------
    # loop on key_rays
    # ----------------

    for ir, kray in enumerate(key_rays):

        # starting points
        ptsx, ptsy, ptsz = coll.get_rays_pts(
            kray,
            segment=segment,
        )
        px = ptsx[0, ...]
        py = ptsy[0, ...]
        pz = ptsz[0, ...]

        # unit vectors
        vx, vy, vz = coll.get_rays_vect(
            kray,
            segment=segment,
        )

        # intersection
        shape = px.shape + px.shape
        interx = np.full(shape, np.nan)
        intery = np.full(shape, np.nan)
        interz = np.full(shape, np.nan)
        for i0, ind0 in enumerate(np.ndindex(px.shape)):
            for i1, ind1 in enumerate(np.ndindex(px.shape)):

                if i1 <= i0:
                    continue

                sli = ind0 + ind1
                interx[sli], intery[sli], interz[sli] = _intersect_line2line(
                    px[ind0], py[ind0], pz[ind0],
                    px[ind1], py[ind1], pz[ind1],
                    vx[ind0], vy[ind0], vz[ind0],
                    vx[ind1], vy[ind1], vz[ind1],
                )

        cx[ir] = np.nanmean(interx)
        cy[ir] = np.nanmean(intery)
        cz[ir] = np.nanmean(interz)

        ux[ir] = np.nanmean(vx)
        uy[ir] = np.nanmean(vy)
        uz[ir] = np.nanmean(vz)

    # --------------
    # take average
    # --------------

    ntot = np.sum(nrays_tot)
    cent = np.r_[
        np.sum(cx * nrays_tot) / ntot,
        np.sum(cy * nrays_tot) / ntot,
        np.sum(cz * nrays_tot) / ntot,
    ]
    nin = np.r_[
        np.sum(ux * nrays_tot) / ntot,
        np.sum(uy * nrays_tot) / ntot,
        np.sum(uz * nrays_tot) / ntot,
    ]

    return cent, nin


def _intersect_line2line(
    px0, py0, pz0,
    px1, py1, pz1,
    vx0, vy0, vz0,
    vx1, vy1, vz1,
):

    # A0M = k0u0
    # A1N = k1u1
    # MN = -A0M + A0A1 + A1N
    # MN = -k0u0 + A0A1 + k1u1
    # MN.u0 = 0 = -k0 + A0A1.u0 + k1(u0.u1)
    # MN.u1 = 0 = -k0(u0.u1) + A0A1.u1 + k1
    #
    # k0 = A0A1.u0 + k1(u0.u1)
    # 0 = k1(1 - (u0.u1)^2) - (A0A1.u0)(u0.u1) + A0A1.u1
    # k1 = ((A0A1.u0)(u0.u1) - A0A1.u1) / (1 - (u0.u1)^2)

    A0A1_x = px1 - px0
    A0A1_y = py1 - py0
    A0A1_z = pz1 - pz0

    A0A1_u0 = A0A1_x * vx0 + A0A1_y * vy0 + A0A1_z * vz0
    A0A1_u1 = A0A1_x * vx1 + A0A1_y * vy1 + A0A1_z * vz1
    u0_u1 = vx0 * vx1 + vy0 * vy1 + vz0 * vz1

    k1 = ((A0A1_u0) * (u0_u1) - A0A1_u1) / (1. - (u0_u1)**2)
    k0 = A0A1_u0 + k1 * u0_u1

    # halfway
    interx = 0.5 * (px0 + k0 * vx0 + px1 + k1 * vx1)
    intery = 0.5 * (py0 + k0 * vy0 + py1 + k1 * vy1)
    interz = 0.5 * (pz0 + k0 * vz0 + pz1 + k1 * vz1)

    return interx, intery, interz


# ########################################################
# ########################################################
#              Get rays angles
# ########################################################


def _get_rays_angles(
    coll=None,
    key_single_pt_cam=None,
    # rays to get angles of
    key_rays=None,
    segment=None,
    # max tolerance
    tol_radius=None,
    # optional indices and convex
    return_indices=None,
    convex_axis=None,
    # verb
    verb=None,
):

    # -------------
    # check inputs
    # -------------

    (
        key_single_pt_cam, key_rays,
        segment, tol_radius,
        return_indices, convex_axis,
        verb,
    ) = _check_rays_angles(
        coll=coll,
        key_single_pt_cam=key_single_pt_cam,
        key_rays=key_rays,
        segment=segment,
        tol_radius=tol_radius,
        return_indices=return_indices,
        convex_axis=convex_axis,
        verb=verb,
    )

    # -------------
    # prepare
    # -------------

    wrays = coll._which_rays
    ref = coll.dobj[wrays][key_rays]['ref'][1:]
    shape = coll.dobj[wrays][key_rays]['shape'][1:]
    kang0, kang1 = coll.dobj[wrays][key_single_pt_cam]['angles']
    units = coll.ddata[kang0]['units']

    # unit vectors
    nin = coll.dobj[wrays][key_single_pt_cam]['nin']
    e0 = coll.dobj[wrays][key_single_pt_cam]['e0']
    e1 = coll.dobj[wrays][key_single_pt_cam]['e1']

    ptsx, ptsy, ptsz = coll.get_rays_pts(key_single_pt_cam)
    cent = np.r_[ptsx[0, 0, 0], ptsy[0, 0, 0], ptsz[0, 0, 0]]

    # -------------
    # get impact factor vs tol_radius
    # -------------

    ptsx, ptsy, ptsz = coll.get_rays_pts(key_rays, segment=segment)
    vx, vy, vz = coll.get_rays_vect(key_rays, segment=segment, norm=True)

    p0c_x = (ptsx[0, ...] - cent[0])
    p0c_y = (ptsy[0, ...] - cent[1])
    p0c_z = (ptsz[0, ...] - cent[2])

    crossx = p0c_y * vz - p0c_z * vy
    crossy = p0c_z * vx - p0c_x * vz
    crossz = p0c_x * vy - p0c_y * vx

    impact = np.sqrt(crossx**2 + crossy**2 + crossz**2)

    # verb
    if verb is True:
        msg = (
            "Angles computation for rays, seen from single_point_camera2d:\n"
            f"\t- single_point_camera2d: {key_single_pt_cam}\n"
            f"\t- key_rays: {key_rays}\n"
            f"\t- max impact: {np.max(impact)} m\n"
        )
        print(msg)

    # -------------
    # compute
    # -------------

    # intialize
    ang0 = np.full(shape, np.nan)
    ang1 = np.full(shape, np.nan)

    # vect from cent
    vx = ptsx[1, ...] - cent[0]
    vy = ptsy[1, ...] - cent[1]
    vz = ptsz[1, ...] - cent[2]
    vnorm = np.sqrt(vx**2 + vy**2 + vz**2)
    vx = vx / vnorm
    vy = vy / vnorm
    vz = vz / vnorm

    # derive angles
    sin1 = vx * e1[0] + vy * e1[1] + vz * e1[2]
    cos1_sin0 = vx * e0[0] + vy * e0[1] + vz * e0[2]
    cos1_cos0 = vx * nin[0] + vy * nin[1] + vz * nin[2]
    cos12 = cos1_sin0**2 + cos1_cos0**2
    cos1 = np.sqrt(cos12)
    ang1 = np.arctan2(sin1, cos1)
    ang0 = np.arctan2(cos1_sin0 / cos1, cos1_cos0 / cos1)

    # adjust units
    if str(units) == 'deg':
        ang0 = ang0 * 180/np.pi
        ang1 = ang1 * 180/np.pi

    # -------------
    # indices
    # -------------

    if return_indices is True:

        bin0 = coll.ddata[kang0]['data']
        dang0 = bin0[1] - bin0[0]
        bin0 = 0.5*(bin0[1:] + bin0[:-1])
        bin0 = np.r_[bin0[0]-dang0, bin0]

        bin1 = coll.ddata[kang0]['data']
        dang1 = bin1[1] - bin1[0]
        bin1 = 0.5*(bin1[1:] + bin1[:-1])
        bin1 = np.r_[bin1[0]-dang1, bin1]

        ind0 = np.searchsorted(bin0, ang0) - 1
        ind1 = np.searchsorted(bin1, ang1) - 1

        # adjust for edges
        ind0[ind0 < 0] = 0
        ind1[ind1 < 0] = 0

        # -------
        # convex_axis

        if convex_axis is not False:

            # shape_angles
            shape_angles = coll.dobj[wrays][key_single_pt_cam]['shape'][1:]

            # shape
            shape_c = tuple([
                ss for ii, ss in enumerate(shape)
                if ii not in convex_axis
            ])

            # unique call
            if len(shape_c) == 0:
                sli = tuple([slice(None) for ii in shape])
                dhull = {
                    (0,): _convexhull(
                        shape_angles=shape_angles,
                        ind0=ind0,
                        ind1=ind1,
                        sli=sli,
                    )
                }

            else:
                # initialize
                dhull = {}
                sli = np.array([
                    slice(None) if ii in convex_axis else 0
                    for ii in range(len(shape))
                ])
                isli = np.array([
                    ii for ii in range(len(shape))
                    if ii not in convex_axis
                ])

                i0 = np.arange(shape_angles[0])
                i1 = np.arange(shape_angles[1])
                i01_full = np.array([
                    np.repeat(i0[:, None], i1.size, axis=1).ravel(),
                    np.repeat(i1[None, :], i0.size, axis=0).ravel(),
                ]).T

                # loop
                for ii, ind in enumerate(np.ndindex(shape_c)):

                    # update slice
                    sli[isli] = ind
                    slii = tuple(sli)

                    # get convex_axis hull
                    dhull[ind] = _convexhull(
                        shape_angles=shape_angles,
                        ind0=ind0,
                        ind1=ind1,
                        sli=slii,
                        i01_full=i01_full,
                    )

    # -------------
    # output
    # -------------

    dout = {
        'angle0': {
            'key': f'{key_rays}_{key_single_pt_cam}_ang0',
            'data': ang0,
            'units': units,
            'ref': ref,
            'dim': 'angle',
        },
        'angle1': {
            'key': f'{key_rays}_{key_single_pt_cam}_ang1',
            'data': ang1,
            'units': units,
            'ref': ref,
            'dim': 'angle',
        },
        'impact': {
            'key': f'{key_rays}_{key_single_pt_cam}_impact',
            'data': impact,
            'ref': ref,
            'units': 'm',
            'dim': 'distance',
        },
        'indices': {},
    }

    if return_indices:
        dout['ind0'] = {
            'key': f'{key_rays}_{key_single_pt_cam}_ind0',
            'data': ind0,
            'ref': ref,
        }
        dout['ind1'] = {
            'key': f'{key_rays}_{key_single_pt_cam}_ind1',
            'data': ind1,
            'ref': ref,
        }

        if convex_axis is not False:
            dout['hull'] = dhull

    return dout


# ########################################################
# ########################################################
#              check inputs rays angles
# ########################################################


def _check_rays_angles(
    coll=None,
    key_single_pt_cam=None,
    # rays to get angles of
    key_rays=None,
    segment=None,
    # max tolerance
    tol_radius=None,
    # optional indices
    return_indices=None,
    convex_axis=None,
    # verb
    verb=None,
):

    # -------------
    # keys
    # -------------

    wrays = coll._which_rays

    # key_rays
    lok = list(coll.dobj.get(wrays, {}).keys())
    key_rays = ds._generic_check._check_var(
        key_rays, 'key_rays',
        types=str,
        allowed=lok,
    )

    # key_single_pt_cam
    lok = [
        k0 for k0 in lok
        if isinstance(coll.dobj[wrays][k0].get('angles'), tuple)
        and len(coll.dobj[wrays][k0]['angles']) == 2
        and all([
            ss in coll.ddata.keys()
            for ss in coll.dobj[wrays][k0].get('angles')
        ])
        and isinstance(coll.dobj[wrays][k0].get('nin'), np.ndarray)
        and isinstance(coll.dobj[wrays][k0].get('e0'), np.ndarray)
        and isinstance(coll.dobj[wrays][k0].get('e1'), np.ndarray)
    ]
    key_single_pt_cam = ds._generic_check._check_var(
        key_single_pt_cam, 'key_single_pt_cam',
        types=str,
        allowed=lok,
    )

    # -----------------
    # segment
    # -----------------

    segment = int(ds._generic_check._check_var(
        segment, 'segment',
        types=(int, float),
        default=0,
    ))

    # -----------------
    # tol_radius
    # -----------------

    if tol_radius is not None:
        tol_radius = float(ds._generic_check._check_var(
            tol_radius, 'tol_radius',
            types=(int, float),
            sign='>0',
        ))

    # -----------------
    # return_indices
    # -----------------

    return_indices = ds._generic_check._check_var(
        return_indices, 'return_indices',
        types=bool,
        default=convex_axis not in [None, False],
    )

    # --------------
    # convex_axis
    # --------------

    if np.isscalar(convex_axis) and not isinstance(convex_axis, bool):
        convex_axis = (int(convex_axis),)

    convex_axis = ds._generic_check._check_var(
        convex_axis, 'convex_axis',
        types=(tuple, bool),
        default=False,
    )

    # case by case
    ndim = len(coll.dobj[wrays][key_rays]['shape'][1:])
    if convex_axis is True:
        convex_axis = tuple(range(ndim))

    elif convex_axis is not False:
        convex_axis = ds._generic_check._check_var_iter(
            convex_axis, 'convex_axis',
            types=tuple,
            types_iter=int,
            allowed=list(range(ndim)) + list(range(-1, -ndim-1, -1)),
        )

        convex_axis = np.array(convex_axis)
        ineg = convex_axis < 0
        convex_axis[ineg] = ndim + convex_axis[ineg]
        convex_axis = tuple(sorted(convex_axis))

    # -----------------
    # verb
    # -----------------

    verb = ds._generic_check._check_var(
        verb, 'verb',
        types=bool,
        default=True,
    )

    return (
        key_single_pt_cam, key_rays,
        segment, tol_radius,
        return_indices, convex_axis,
        verb,
    )


# ####################################################
# ####################################################
#            get ConvexHull
# ####################################################


def _convexhull(
    shape_angles=None,
    ind0=None,
    ind1=None,
    sli=None,
    i01_full=None,
):
    # ---------------
    # initialize
    # ---------------

    ind = np.zeros(shape_angles, dtype=bool)

    # ---------------
    # get convex hull
    # ---------------

    ui0 = np.unique(ind0[sli])
    ui1 = np.unique(ind1[sli])
    if ui0.size == 1:
        npts = ui1[-1] - ui1[0] + 1
        slii = (
            np.full((npts,), ui0[0]),
            np.arange(ui1[0], min(ui1[-1]+1, ind.shape[1])),
        )
        ind[slii] = True

    elif ui1.size == 1:
        npts = ui0[-1] - ui0[0] + 1
        slii = (
            np.arange(ui0[0], min(ui0[-1]+1, ind.shape[0])),
            np.full((npts,), ui1[0]),
        )
        ind[slii] = True

    # ---------------
    # get convex hull
    # ---------------

    else:
        i0_flat = ind0[sli].ravel()
        i1_flat = ind1[sli].ravel()
        hull = scpsp.ConvexHull(
            np.array([i0_flat, i1_flat]).T
        )

        # get indices of angles in hull
        path = Path(
            np.array([i0_flat[hull.vertices], i1_flat[hull.vertices]]).T
        )
        ind = path.contains_points(i01_full).reshape(shape_angles)

    return ind



# -*- coding: utf-8 -*-


import copy


import numpy as np
import datastock as ds


import datetime as dtm      # DB


from . import _class8_vos_broadband as _vos_broadband
from . import _class8_vos_spectro as _vos_spectro
from . import _class8_los_angles


# ###############################################################
# ###############################################################
#                       Main
# ###############################################################


def compute_vos(
    # resources
    coll=None,
    key_diag=None,
    key_cam=None,
    key_mesh=None,
    config=None,
    # parameters
    res_RZ=None,
    res_phi=None,
    lamb=None,
    res_lamb=None,
    res_rock_curve=None,
    n0=None,
    n1=None,
    convexHull=None,
    # margins
    margin_poly=None,
    # user-defined limits
    user_limits=None,
    # keep3d
    keep3d=None,
    return_vector=None,
    # options
    add_points=None,
    # spectro-only
    rocking_curve_fw=None,
    rocking_curve_max=None,
    # bool
    visibility=None,
    convex=None,
    check=None,
    verb=None,
    debug=None,
    # storing
    store=None,
    overwrite=None,
    replace_poly=None,
    timing=None,
):

    if timing:
        t0 = dtm.datetime.now()     # DB

    # ------------
    # check inputs

    (
        key_diag,
        key_mesh,
        spectro,
        is2d,
        doptics,
        dcompute,
        res_RZ,
        res_phi,
        res_lamb,
        keep3d,
        return_vector,
        convexHull,
        visibility,
        verb,
        debug,
        store,
        overwrite,
        timing,
    ) = _check(
        # resources
        coll=coll,
        key_diag=key_diag,
        key_cam=key_cam,
        key_mesh=key_mesh,
        # parameters
        res_RZ=res_RZ,
        res_phi=res_phi,
        lamb=lamb,
        res_lamb=res_lamb,
        convexHull=convexHull,
        # bool
        keep3d=keep3d,
        return_vector=return_vector,
        visibility=visibility,
        verb=verb,
        debug=debug,
        store=store,
        overwrite=overwrite,
        timing=timing,
    )

    # ----------
    # verb

    if verb is True:
        msg = f"\nComputing vos for diag '{key_diag}':"
        print(msg)

    # -----------
    # prepare

    if spectro:
        func = _vos_spectro._vos
    else:
        func = _vos_broadband._vos

    # ------------
    # sample mesh

    dsamp = coll.get_sample_mesh(
        key=key_mesh,
        res=res_RZ,
        mode='abs',
        grid=True,
        in_mesh=True,
        # non-used
        x0=None,
        x1=None,
        Dx0=None,
        Dx1=None,
        imshow=False,
        store=False,
        kx0=None,
        kx1=None,
    )

    sh = dsamp['x0']['data'].shape
    x0f = dsamp['x0']['data'].ravel()
    x1f = dsamp['x1']['data'].ravel()

    sh1 = tuple([ss + 2 for ss in sh])
    bool_cross = np.zeros(sh1, dtype=bool)

    x0u = dsamp['x0']['data'][:, 0]
    x1u = dsamp['x1']['data'][0, :]
    x0l = np.r_[x0u[0] - (x0u[1] - x0u[0]), x0u, x0u[-1] + (x0u[-1] - x0u[-2])]
    x1l = np.r_[x1u[0] - (x1u[1] - x1u[0]), x1u, x1u[-1] + (x1u[-1] - x1u[-2])]
    x0l = np.repeat(x0l[:, None], x1l.size, axis=1)
    x1l = np.repeat(x1l[None, :], x0l.shape[0], axis=0)

    dx0 = x0u[1] - x0u[0]
    dx1 = x1u[1] - x1u[0]

    # --------------
    # prepare optics

    doptics = coll._dobj['diagnostic'][key_diag]['doptics']

    # -------------
    # user-defined

    user_limits = _get_user_limits(
        coll=coll,
        user_limits=user_limits,
        doptics=doptics,
        key_cam=list(dcompute.keys()),
        # lims
        x0u=x0u,
        x1u=x1u,
    )


    # timing
    if timing:
        t1 = dtm.datetime.now()     # DB
        dt1 = (t1 - t0).total_seconds()
    dt11, dt22 = 0, 0
    dt111, dt222, dt333 = 0, 0, 0
    dt1111, dt2222, dt3333, dt4444 = 0, 0, 0, 0

    # --------------
    # prepare optics

    dvos, dref = {}, {}
    for k0 in dcompute.keys():

            # ------------------
            # call relevant func

            (
                dvos[k0], dref[k0],
                dt11, dt22,
                dt111, dt222, dt333,
                dt1111, dt2222, dt3333, dt4444,
            ) = func(
                # ressources
                coll=coll,
                doptics=doptics,
                key_diag=key_diag,
                key_cam=k0,
                dsamp=dsamp,
                # inputs sample points
                x0u=x0u,
                x1u=x1u,
                x0f=x0f,
                x1f=x1f,
                x0l=x0l,
                x1l=x1l,
                dx0=dx0,
                dx1=dx1,
                # options
                sh=sh,
                res_RZ=res_RZ,
                res_phi=res_phi,
                lamb=lamb,
                res_lamb=res_lamb,
                res_rock_curve=res_rock_curve,
                n0=n0,
                n1=n1,
                convexHull=convexHull,
                bool_cross=bool_cross,
                # user-defined limits
                user_limits=user_limits,
                # keep3d
                keep3d=keep3d,
                return_vector=return_vector,
                # parameters
                margin_poly=margin_poly,
                config=config,
                visibility=visibility,
                verb=verb,
                # debug
                debug=debug,
                # timing
                timing=timing,
                dt11=dt11,
                dt111=dt111,
                dt1111=dt1111,
                dt2222=dt2222,
                dt3333=dt3333,
                dt4444=dt4444,
                dt222=dt222,
                dt333=dt333,
                dt22=dt22,
            )

            dvos[k0]['keym'] = key_mesh
            dvos[k0]['res_RZ'] = res_RZ
            dvos[k0]['res_phi'] = res_phi
            if spectro is True:
                dvos[k0]['res_lamb'] = res_lamb
                dvos[k0]['res_rock_curve'] = res_rock_curve

    # timing
    if timing:
        t2 = dtm.datetime.now()     # DB
        print("\nTIMING\n--------")
        print(f'Prepare: {dt1} s')
        print(f'\tdt11 (pepare cam): {dt11} s')
        print(f'\t\tdt111 (prepare pix): {dt111} s')
        print(f"\t\t\tdt1111 (prepare): {dt1111} s")
        print(f"\t\t\tdt2222 (compute): {dt2222} s")
        print(f"\t\t\tdt3333 (format):  {dt3333} s")
        print(f"\t\t\tdt4444 (ind_bool):  {dt4444} s")
        print(f'\t\tdt222 (compute): {dt222} s')
        print(f'\t\tdt333 (get poly): {dt333} s')
        print(f'\tdt22 (interp poly): {dt22} s')
        print(f'loop total: {(t2-t1).total_seconds()} s')

    # -------------
    # replace

    if store is True:

        _store(
            coll=coll,
            key_diag=key_diag,
            dvos=dvos,
            dref=dref,
            spectro=spectro,
            overwrite=overwrite,
            replace_poly=replace_poly,
        )

    return dvos, dref


# ###########################################################
# ###########################################################
#               check
# ###########################################################


def _check(
    coll=None,
    key_diag=None,
    key_cam=None,
    key_mesh=None,
    res_RZ=None,
    res_phi=None,
    lamb=None,
    res_lamb=None,
    convexHull=None,
    # keep3d
    keep3d=None,
    return_vector=None,
    # bool
    visibility=None,
    check=None,
    verb=None,
    debug=None,
    store=None,
    overwrite=None,
    timing=None,
):

    # --------
    # key_diag

    lok = [
        k0 for k0, v0 in coll.dobj.get('diagnostic', {}).items()
        if any([len(v1['optics']) > 0 for v1 in v0['doptics'].values()])
    ]
    key_diag = ds._generic_check._check_var(
        key_diag, 'key_diag',
        types=str,
        allowed=lok,
    )

    # spectro, is2d
    spectro = coll.dobj['diagnostic'][key_diag]['spectro']
    is2d = coll.dobj['diagnostic'][key_diag]['is2d']

    # doptics
    doptics = coll.dobj['diagnostic'][key_diag]['doptics']

    # ------------
    # key_cam

    lok = [k0 for k0, v0 in doptics.items() if len(v0['optics']) > 0]
    if isinstance(key_cam, str):
        key_cam = [key_cam]
    key_cam = ds._generic_check._check_var_iter(
        key_cam, 'key_cam',
        types=list,
        types_iter=str,
        allowed=lok,
    )

    # -----------------
    # doptics, dcompute

    doptics = {k0: v0 for k0, v0 in doptics.items() if k0 in key_cam}

    dcompute = {
        k0: {'compute': len(v0['optics']) > 0}
        for k0, v0 in doptics.items()
    }

    # --------
    # key_mesh

    wm = coll._which_mesh
    lok = list(coll.dobj.get(wm, {}).keys())
    key_mesh = ds._generic_check._check_var(
        key_mesh, 'key_mesh',
        types=str,
        allowed=lok,
    )

    # -------------------------------------------------
    # ldeti: list of individual camera dict (per pixel)

    for k0, v0 in doptics.items():

        dgeom = coll.dobj['camera'][k0]['dgeom']
        cx, cy, cz = coll.get_camera_cents_xyz(key=k0)
        dvect = coll.get_camera_unit_vectors(key=k0)
        outline = dgeom['outline']
        out0 = coll.ddata[outline[0]]['data']
        out1 = coll.ddata[outline[1]]['data']
        is2d = dgeom['nd'] == '2d'
        par = dgeom['parallel']
        dcompute[k0]['shape0'] = cx.shape

        if is2d:
            cx = cx.ravel()
            cy = cy.ravel()
            cz = cz.ravel()

        nd = cx.size

        dcompute[k0]['ldet'] = [
            {
                'cents_x': cx[ii],
                'cents_y': cy[ii],
                'cents_z': cz[ii],
                'outline_x0': out0,
                'outline_x1': out1,
                'nin_x': dvect['nin_x'] if par else dvect['nin_x'][ii],
                'nin_y': dvect['nin_y'] if par else dvect['nin_y'][ii],
                'nin_z': dvect['nin_z'] if par else dvect['nin_z'][ii],
                'e0_x': dvect['e0_x'] if par else dvect['e0_x'][ii],
                'e0_y': dvect['e0_y'] if par else dvect['e0_y'][ii],
                'e0_z': dvect['e0_z'] if par else dvect['e0_z'][ii],
                'e1_x': dvect['e1_x'] if par else dvect['e1_x'][ii],
                'e1_y': dvect['e1_y'] if par else dvect['e1_y'][ii],
                'e1_z': dvect['e1_z'] if par else dvect['e1_z'][ii],
            }
            for ii in range(nd)
        ]

    # -----------
    # res_RZ

    if res_RZ is None:
        res_RZ = 0.01
    if np.isscalar(res_RZ):
        res_RZ = np.r_[res_RZ, res_RZ]
    res_RZ = np.atleast_1d(res_RZ).ravel().astype(float)
    assert res_RZ.size == 2
    res_RZ = res_RZ.tolist()

    # -----------
    # res_phi

    if res_phi is None:
        res_phi = 0.01

    # -----------
    # res_lamb

    if res_lamb is None and lamb is None:
        res_lamb = 0.01-10

    # -----------
    # keep3d

    keep3d = ds._generic_check._check_var(
        keep3d, 'keep3d',
        types=bool,
        default=False,
    )

    # -----------
    # return_vector

    return_vector = ds._generic_check._check_var(
        return_vector, 'return_vector',
        types=bool,
        default=False,
    )

    # -----------
    # convexHull - to get overall pcross and phor, faster if many pixels

    convexHull = ds._generic_check._check_var(
        convexHull, 'convexHull',
        types=bool,
        default=False,
    )

    # -----------
    # visibility

    visibility = ds._generic_check._check_var(
        visibility, 'visibility',
        types=bool,
        default=True,
    )

    # -----------
    # verb

    verb = ds._generic_check._check_var(
        verb, 'verb',
        types=bool,
        default=True,
    )

    # -----------
    # debug

    if debug is None:
        debug = False

    if callable(debug):
        pass
    else:
        debug = ds._generic_check._check_var(
            debug, 'debug',
            types=bool,
            default=False,
        )

    # -----------
    # store

    store = ds._generic_check._check_var(
        store, 'store',
        types=bool,
        default=False,
    )

    # -----------
    # overwrite

    overwrite = ds._generic_check._check_var(
        overwrite, 'overwrite',
        types=bool,
        default=False,
    )

    # -----------
    # timing

    timing = ds._generic_check._check_var(
        timing, 'timing',
        types=bool,
        default=False,
    )

    return (
        key_diag,
        key_mesh,
        spectro,
        is2d,
        doptics,
        dcompute,
        res_RZ,
        res_phi,
        res_lamb,
        keep3d,
        return_vector,
        convexHull,
        visibility,
        verb,
        debug,
        store,
        overwrite,
        timing,
    )


def _user_limits_err(user_limits, lc=None):
    msg = (
        "user_limits must be either:\n"
        "\t- None: not used\n"
        "\t- True: remove all limits\n"
        "\t- dict: specify limits with at least one of the keys:\n"
        "\t\t- 'DR': list of 2 scalars or None\n"
        "\t\t- 'DZ': list of 2 scalars or None\n"
        "\t\t- 'Dphi': list of 2 scalars or None\n"
        + ("" if lc is None else f"lc = {lc}\n")
        + f"Provided:\n{user_limits}\n"
    )
    raise Exception(msg)



def _get_user_limits(
    coll=None,
    user_limits=None,
    doptics=None,
    key_cam=None,
    # lims
    x0u=None,
    x1u=None,
):

    # ---------------
    # check

    # default
    c0 = all([v0['los'] is None for v0 in doptics.values()])
    if user_limits is None and c0:
        user_limits = True

    # dict
    if user_limits not in [None, True]:

        if not isinstance(user_limits, dict):
            _user_limits_err(user_limits)

        lc = [
            user_limits.get(ss) is not None
            and len(user_limits[ss]) == 2
            for ss in ['DR', 'DZ', 'Dphi']
        ]

        if not any(lc):
            _user_limits_err(user_limits, lc)

    # --------------
    # trivial

    if user_limits is None:
        return

    elif user_limits is True:
        user_limits = {}

    # --------------------
    # Preliminary check on Dphi
    # --------------------

    if user_limits.get('Dphi') is not None:
        user_limits['Dphi'] = ds._generic_check._check_flat1darray(
            user_limits['Dphi'], "user_limits.get('Dphi')",
            dtype=float,
            size=2,
            unique=True,
        )
        phi = np.linspace(user_limits['Dphi'][0], user_limits['Dphi'][1], 50)

    # --------------------
    # user-defined pcross
    # --------------------

    if user_limits.get('DR') is not None or user_limits.get('DZ') is not None:

        # DR
        if user_limits.get('DR') is None:
            user_limits['DR'] = [x0u[0]-1e-9, x0u[-1]+1e-9]
        user_limits['DR'] = ds._generic_check._check_flat1darray(
            user_limits['DR'], "user_limits.get('DR')",
            dtype=float,
            size=2,
            unique=True,
            sign='>=0',
        )

        # DZ
        if user_limits.get('DZ') is None:
            user_limits['DZ'] = [x1u[0]-1e-9, x1u[-1]+1e-9]
        user_limits['DZ'] = ds._generic_check._check_flat1darray(
            user_limits['DZ'], "user_limits['DZ']",
            dtype=float,
            size=2,
            unique=True,
        )

        # pcross_user
        DR = user_limits['DR']
        DZ = user_limits['DZ']
        user_limits['pcross_user'] = np.array([
            np.r_[DR[0], DR[1], DR[1], DR[0]],
            np.r_[DZ[0], DZ[0], DZ[1], DZ[1]],
        ])

        # --------------------
        # user-defined phor
        # --------------------

        # Dphi
        # phor_user
        if user_limits.get('Dphi') is not None:
            user_limits['phor_user'] = np.array([
                DR[1] * np.r_[-1, 1, 1, -1],
                DR[1] * np.r_[-1, -1, 1, 1],
            ])
        else:

            user_limits['phor_user'] = np.array([
                np.r_[
                    DR[0]*np.cos(phi),
                    DR[1]*np.cos(phi[::-1]),
                ],
                np.r_[
                    DR[0]*np.sin(phi),
                    DR[1]*np.sin(phi[::-1]),
                ],
            ])

    # -------------------
    # case with only Dphi => use each pixel's pcross
    # -------------------

    elif user_limits.get('Dphi') is not None:

        # Give phor_user the proper shape like phor TBF
        user_limits['phor0'] = {}
        user_limits['phor1'] = {}
        user_limits['dphi'] = {}

        for kcam in key_cam:
            kpc0, kpc1 = doptics[kcam]['dvos']['pcross']

            shape = coll.ddata[kpc0]['data'].shape
            pcross0 = coll.ddata[kpc0]['data']
            pcross1 = coll.ddata[kpc1]['data']

            R = np.hypot(pcross0, pcross1)
            Rmin = np.min(R, axis=0)
            Rmax = np.max(R, axis=0)

            phi = phi.reshape(tuple([phi.size] + [1]*Rmin.ndim))
            phor0 = np.concatenate(
                (
                    Rmin[None, ...] * np.cos(phi),
                    Rmax[None, ...] * np.cos(phi[::-1]),
                ),
                axis=0,
            )

            phor1 = np.concatenate(
                (
                    Rmin[None, ...] * np.sin(phi),
                    Rmax[None, ...] * np.sin(phi[::-1]),
                ),
                axis=0,
            )

            user_limits['phor0'][kcam] = phor0
            user_limits['phor1'][kcam] = phor1

            shape_cam = coll.dobj['camera'][kcam]['dgeom']['shape']
            user_limits['dphi'][kcam] = np.array([
                np.full(shape_cam, user_limits['Dphi'][0]),
                np.full(shape_cam, user_limits['Dphi'][1]),
            ])

    # -----------
    # clean-up
    # ----------

    if len(user_limits) == 0:
        user_limits = None

    return user_limits


# ###########################################################
# ###########################################################
#               store
# ###########################################################


def _store(
    coll=None,
    key_diag=None,
    dvos=None,
    dref=None,
    spectro=None,
    overwrite=None,
    replace_poly=None,
):

    # ------------
    # check inputs

    replace_poly = ds._generic_check._check_var(
        replace_poly, 'replace_poly',
        types=bool,
        default=True,
    )

    # ----------------------
    # prepare what to store

    lk_com = [
        'indr_cross', 'indz_cross',
        'indr_3d', 'indz_3d', 'phi_3d',
        'vectx_3d', 'vecty_3d', 'vectz_3d',
    ]

    if spectro is True:
        lk = [
            'lamb',
            'ph', 'cos', 'ncounts',
            'phi_min', 'phi_max',
            # optional
            'phi_mean',
            'dV', 'etendlen',
        ]
    else:
        lk = ['sang_cross', 'ang_pol_cross', 'ang_tor_cross', 'sang_3d']


    # ------------
    # store

    doptics = coll._dobj['diagnostic'][key_diag]['doptics']

    for k0, v0 in dvos.items():

        # ----------------
        # pcross replacement

        if replace_poly and v0.get('pcross0') is not None:

            if v0.get('phor0') is None:
                phor0, phor1 = None, None
            else:
                phor0, phor1 = v0['phor0']['data'], v0['phor1']['data']

            # re-use previous keys
            if doptics[k0].get('dvos') is None:

                _class8_los_angles._vos_from_los_store(
                    coll=coll,
                    key=key_diag,
                    key_cam=k0,
                    pcross0=v0['pcross0']['data'],
                    pcross1=v0['pcross1']['data'],
                    phor0=phor0,
                    phor1=phor1,
                    dphi=None,
                )

            else:
                kpc0, kpc1 = doptics[k0]['dvos']['pcross']
                kr = coll.ddata[kpc0]['ref'][0]

                # safety check
                shape_pcross = v0['pcross0']['data'].shape
                if coll.ddata[kpc0]['data'].shape[1:] != shape_pcross[1:]:
                    msg = "Something is wrong"
                    raise Exception(msg)

                coll._dref[kr]['size'] = shape_pcross[0]
                coll._ddata[kpc0]['data'] = v0['pcross0']['data']
                coll._ddata[kpc1]['data'] = v0['pcross1']['data']
                if phor0 is not None:
                    kph0, kph1 = doptics[k0]['dvos']['phor']
                    coll._ddata[kph0]['data'] = v0['phor0']['data']
                    coll._ddata[kph1]['data'] = v0['phor1']['data']

        # ----------------
        # add ref of sang

        for k1, v1 in dref[k0].items():
            if v1['key'] in coll.dref.keys():
                if overwrite is True:
                    coll.remove_ref(v1['key'], propagate=True)
                    coll.add_ref(**v1)
                elif v1['size'] != coll.dref[v1['key']]['size']:
                    msg = (
                        f"Mismatch between new vs existing size ref {k1} '{v1['key']}'"
                        f"\t- existing size = {coll.dref[k1]['size']}\n"
                        f"\t- new size      = {v1['size']}\n"
                    )
                    raise Exception(msg)
                else:
                    pass
            else:
                coll.add_ref(**v1)

        # ----------------
        # add data

        for k1 in lk_com + lk:

            if k1 not in v0.keys():
                continue

            if v0[k1]['key'] in coll.ddata.keys():
                if overwrite is True:
                    coll.remove_data(key=v0[k1]['key'])
                else:
                    msg = (
                        f"Not overwriting existing data '{v0[k1]['key']}'\n"
                        "To force update use overwrite = True"
                    )
                    raise Exception(msg)

            coll.add_data(**v0[k1])

        # ---------------
        # add in doptics

        doptics[k0]['dvos']['keym'] = v0['keym']
        doptics[k0]['dvos']['res_RZ'] = v0['res_RZ']
        doptics[k0]['dvos']['res_phi'] = v0['res_phi']
        doptics[k0]['dvos']['ind_cross'] = (v0['indr_cross']['key'], v0['indz_cross']['key'])

        # 3d
        doptics[k0]['dvos']['indr_3d'] = v0.get('indr_3d', {}).get('key')
        doptics[k0]['dvos']['indz_3d'] = v0.get('indz_3d', {}).get('key')
        doptics[k0]['dvos']['phi_3d'] = v0.get('phi_3d', {}).get('key')
        doptics[k0]['dvos']['sang_3d'] = v0.get('sang_3d', {}).get('key')

        # vect
        doptics[k0]['dvos']['vectx_3d'] = v0.get('vectx_3d', {}).get('key')
        doptics[k0]['dvos']['vecty_3d'] = v0.get('vecty_3d', {}).get('key')
        doptics[k0]['dvos']['vectz_3d'] = v0.get('vectz_3d', {}).get('key')

        # spectro
        if spectro:
            doptics[k0]['dvos']['res_lamb'] = v0['res_lamb']
            doptics[k0]['dvos']['res_rock_curve'] = v0['res_rock_curve']

        # -----------------
        # add data keys to doptics

        for k1 in lk:
            if k1 in v0.keys():
                doptics[k0]['dvos'][k1] = v0[k1]['key']


# ###############################################################
# ###############################################################
#                       get / check
# ###############################################################


def _check_get_dvos(
    coll=None,
    key=None,
    key_cam=None,
    dvos=None,
):

    # ------------
    # keys

    key_diag, key_cam = coll.get_diagnostic_cam(
        key=key,
        key_cam=key_cam,
    )
    spectro = coll.dobj['diagnostic'][key_diag]['spectro']

    # -------------------
    # prepare keys

    lk_sca = ['res_RZ', 'res_phi']
    if spectro is True:
        lk_sca += ['res_lamb', 'res_rock_curve']
        lk = [
            'lamb',
            'phi_min', 'phi_max', 'phi_mean',
            'ph', 'ncounts', 'cos',
            'dV', 'etendlen',
        ]
        lk_opt = []

    else:
        lk = [
            'sang_cross',
        ]

        lk_opt = [
            'sang_3d',
            'indr_3d', 'indz_3d', 'phi_3d',
            'vectx_3d', 'vecty_3d', 'vectz_3d'
        ]

    lkcom = ['keym', 'indr_cross', 'indz_cross']

    lk_all = lk_sca + lk + lkcom

    # ------
    # dvos

    if dvos is None:

        dvos = {}
        doptics = coll.dobj['diagnostic'][key_diag]['doptics']
        for k0 in key_cam:

            # safety check 1
            if doptics[k0].get('dvos') is None:
                msg = (
                    "Please provide dvos if coll.dobj['diagnostic']"
                    f"['{key_diag}']['{k0}']['doptics']['dvos'] is None!"
                )
                raise Exception(msg)

            dop = doptics[k0]['dvos']

            # safety check 2
            if dop.get('keym') is None:
                msg = (
                    "dvos was neither pre-computed nor provided for:\n"
                    f"\t- diag: '{key_diag}'\n"
                    f"\t- cam:  '{k0}'"
                )
                raise Exception(msg)

            # fill in dict with mesh and indices
            dvos[k0] = {
                'keym': dop['keym'],
                'indr_cross': coll.ddata[dop['ind_cross'][0]],
                'indz_cross': coll.ddata[dop['ind_cross'][1]],
            }

            # fill in with res
            for k1 in lk_sca:
                dvos[k0][k1] = dop[k1]

            # fill in with the rest
            for k1 in lk:
                if k1 in dop.keys():
                    dvos[k0][k1] = {
                        'key': dop[k1],
                        **coll.ddata[dop[k1]],
                    }
        isstore = True

    else:
        isstore = False

    # copy to avoid changing the original
    dvos = copy.deepcopy(dvos)

    # ------------------
    # check keys of dvos

    # check
    c0 = (
        isinstance(dvos, dict)
        and all([
            k0 in dvos.keys()
            and all([k1 in dvos[k0].keys() for k1 in lk_all])
            for k0 in key_cam
        ])
    )

    # raise exception
    if not c0:
        msg = (
            "Arg dvos must be a dict with, for each camera, the keys:\n"
            f"\t- expected: {sorted(lk_all)}"
        )
        if isinstance(dvos, dict):
            k0 = list(dvos.keys())[0]
            if isinstance(dvos[k0], dict):
                lkout = [k1 for k1 in lk_all if k1 not in dvos[k0].keys()]
                msg += f"\n\t- dvos['{k0}'] missing: {lkout}"
        raise Exception(msg)

    # only keep desired cams
    lkout = [k0 for k0 in dvos.keys() if k0 not in key_cam]
    for k0 in lkout:
        del dvos[k0]

    return key_diag, dvos, isstore


# ###############################################################
# ###############################################################
#                       get vos to 3d
# ###############################################################


def get_dvos_xyz(coll=None, key_diag=None, key_cam=None, dvos=None):

    # ---------
    # get dvos

    key_diag, dvos, isstore = coll.check_diagnostic_dvos(
        key=key_diag,
        key_cam=key_cam,
        dvos=dvos,
    )

    # check
    k3d = 'indr_3d'
    if not all([v0.get(k3d) is not None for v0 in dvos.values()]):
        lstr = [
            f"\t\t- dvos['{k0}']: '{k3d}' {'not' if v0.get(k3d) is None else ''} available"
            for k0, v0 in dvos.items()
        ]
        msg = (
            "dvos can only provide (x, y, z) if it contains 3d information!\n"
            f"\t- key_diag: {key_diag}\n"
            + "\n".join(lstr)
        )
        raise Exception(msg)

    # ---------
    # to xyz

    for k0, v0 in dvos.items():

        # mesh
        keym = v0['keym']

        # sample
        dsamp = coll.get_sample_mesh(
            key=v0['keym'],
            res=v0['res_RZ'],
            mode='abs',
            grid=False,
            in_mesh=True,
            # non-used
            x0=None,
            x1=None,
            Dx0=None,
            Dx1=None,
            imshow=False,
            store=False,
            kx0=None,
            kx1=None,
        )

        # R, Z
        x0u = dsamp['x0']['data']
        x1u = dsamp['x1']['data']

        # store
        ref = v0['indr_3d']['ref']
        dvos[k0].update({
            'ptsx_3d': {
                'data': x0u[v0['indr_3d']['data']] * np.cos(v0['phi_3d']['data']),
                'ref': ref,
                'units': 'm',
            },
            'ptsy_3d': {
                'data': x0u[v0['indr_3d']['data']] * np.sin(v0['phi_3d']['data']),
                'ref': ref,
                'units': 'm',
            },
            'ptsz_3d': {
                'data': x1u[v0['indz_3d']['data']],
                'ref': ref,
                'units': 'm',
            },
        })

    return dvos




import os
import json


import datastock as ds


from ..data import Collection
from . import _ddef
from . import _common
from . import _equilibrium


# ################################################
# ################################################
#              DEFAULT
# ################################################


_IDS_ORDER = [
    'summary',
    'pulse_schedule',
    'equilibrium',
    # 'core_profiles',
]


_DEXTRACT = {
    'summary': [_common.main],
    'pulse_schedule': [_common.main],
    'equilibrium': [_common.main, _equilibrium.main],
    'core_profiles': [_common.main],
}


# ################################################
# ################################################
#              Main
# ################################################


def load_from_omas(
    # resources
    pfe=None,
    coll=None,
    # ids to include
    ids=None,
    # options
    prefix=None,
    strict=None,
    dshort=None,
    warn=None,
):

    # ------------
    # check inputs
    # ------------

    pfe, coll, lids, prefix, dshort, warn = _check(
        pfe=pfe,
        coll=coll,
        ids=ids,
        prefix=prefix,
        dshort=dshort,
        warn=warn,
    )

    # Collection
    if coll is None:
        coll = Collection()

    # -------------
    # load json and check
    # -------------

    dout = _load_json(pfe)

    # -------------
    # extract data to coll
    # -------------

    ids_order = [
        ids for ids in _IDS_ORDER
        if ids in dout.keys()
        and ids in lids
    ]

    # loop to extract
    for ids in ids_order:
        lfunc = _DEXTRACT.get(ids)
        if lfunc is not None:
            for func in lfunc:
                func(
                    din=dout,
                    coll=coll,
                    ids=ids,
                    prefix=prefix,
                    dshort=dshort,
                    strict=strict,
                    warn=warn,
                )

    return coll


# ###########################################################
# ###########################################################
#              Check
# ###########################################################


def _check(
    pfe=None,
    coll=None,
    ids=None,
    prefix=None,
    dshort=None,
    warn=None,
):

    # -------------
    # dshort
    # -------------

    if dshort is None:
        dshort = _ddef.get_dshort()

    if not isinstance(dshort, dict):
        msg = (
            "Arg dshort must be a dict\n"
            f"Provided:\n{dshort}\n"
        )
        raise Exception(msg)

    # --------------
    # pfe
    # --------------

    if not isinstance(pfe, dict):
        c0 = (
            isinstance(pfe, str)
            and os.path.isfile(pfe)
            and pfe.endswith('.json')
        )
        if not c0:
            msg = (
                "Arg 'pfe' must be a path/file.ext to a valid json file!\n"
                f"Provided:\n{pfe}\n"
            )
            raise Exception(msg)

    # --------------
    # coll
    # --------------

    if coll is not None:
        if not isinstance(coll, Collection):
            msg = (
                "Arg coll must be a tf.data.Collection instance!\n"
                f"Provided:\n{coll}\n"
            )
            raise Exception(msg)

    # --------------
    # ids
    # --------------

    if isinstance(ids, str):
        ids = [ids]
    ids = ds._generic_check._check_var_iter(
        ids, 'ids',
        types=(list, tuple),
        types_iter=str,
        allowed=_IDS_ORDER,
        default=_IDS_ORDER,
    )

    # --------------
    # prefix
    # --------------

    if prefix is None:
        if isinstance(pfe, str):
            prefix = os.path.split(pfe)[1].strip('.json')
        else:
            prefix = ''

    if not isinstance(prefix, str):
        msg = f"Arg key must be a str!\nProvided: {prefix}\n"
        raise Exception(msg)

    # --------------
    # warn
    # --------------

    warn = ds._generic_check._check_var(
        warn, 'warn',
        types=bool,
        default=True,
    )

    return pfe, coll, ids, prefix, dshort, warn


# ###########################################################
# ###########################################################
#              load json
# ###########################################################


def _load_json(
    pfe=None,
    coll=None,
    key=None,
):

    # --------------
    # load
    # --------------

    if isinstance(pfe, str):
        with open(pfe) as ff:
            dout = json.load(ff)
    else:
        dout = pfe

    # --------------
    # basic checks on content
    # --------------

    for k0 in _DEXTRACT.keys():
        if dout.get(k0) is not None:
            if not isinstance(dout[k0], dict):
                msg = (
                    "File must contain nested dict!\n"
                    f"pfe = {pfe}\n"
                )
                raise Exception(msg)

    return dout



# -*- coding: utf-8 -*-


# Built-in
import copy


# tofu
# from tofu import __version__ as __version__
from ._class00_Config import Config as Previous
from . import _class01_compute as _compute
from . import _class01_load_equilibrium as _load_equilibrium


__all__ = ['Plasma2D']


_WHICH_MESH = 'mesh'
_QUANT_R = 'R'
_QUANT_Z = 'Z'


# #############################################################################
# #############################################################################
#                           Plasma2D
# #############################################################################


class Plasma2D(Previous):

    _ddef = copy.deepcopy(Previous._ddef)

    # _show_in_summary_core = ['shape', 'ref', 'group']
    _dshow = dict(Previous._dshow)

    # _quant_R = _QUANT_R
    # _quant_Z = _QUANT_Z

    # -------------------
    # units conversione
    # -------------------

    def convert_units_spectral(
        self,
        data=None,
        units=None,
        units_in=None,
    ):

        return _compute.convert_spectral_units(
            coll=self,
            data=data,
            units=units,
            units_in=units_in,
        )

    # -------------------
    # load specific file formats
    # -------------------

    def load_equilibrium_from_files(
        self,
        dpfe=None,
        returnas=None,
        # keys
        kmesh=None,
        # user-defined dunits
        dunits=None,
        # group naming
        func_key_groups=None,
        # sorting
        sort_vs=None,
        # options
        verb=None,
        strict=None,
        explore=None,
    ):
        """ Load an equilibria maps from several files

        Arg dpfe is fed to ds.get_files()
        Can be a dict of path / list of patterns
        Handles either:
            - .eqdsk
            - .mat (mep files)

        If all files do not have the same mesh size, they are grouped by shape
            - func_key_groups: callable to name those groups

        Loaded files can be  sorted vs any scalar quantity using
            - sort_vs

        Loads the R, Z mesh and corresponding 2d maps
        Also loads all 1d or scalar quantities
        """
        return _load_equilibrium.main(
            dpfe=dpfe,
            returnas=returnas,
            # keys
            kmesh=kmesh,
            # user-defined dunits
            dunits=dunits,
            # group naming
            func_key_groups=func_key_groups,
            # sorting
            sort_vs=sort_vs,
            # options
            verb=verb,
            strict=strict,
            explore=explore,
        )

    # -------------------
    # get data time
    # -------------------

    def get_time(
        self,
        key=None,
        t=None,
        indt=None,
        ind_strict=None,
        dim=None,
    ):
        """ Return the time vector or time macthing indices

        hastime, keyt, reft, keyt, val, dind = self.get_time(key='prof0')

        Return
        ------
        hastime:    bool
            flag, True if key has a time dimension
        keyt:       None /  str
            if hastime and a time vector exists, the key to that time vector
        t:          None / np.ndarray
            if hastime
        dind:       dict, with:
            - indt:  None / np.ndarray
                if indt or t was provided, and keyt exists
                int indices of nearest matching times
            - indtu: None / np.ndarray
                if indt is returned, np.unique(indt)
            - indtr: None / np.ndarray
                if indt is returned, a bool (ntu, nt) array
            - indok: None / np.ndarray
                if indt is returned, a bool (nt,) array

        """

        if dim is None:
            dim = 'time'

        return self.get_ref_vector(
            key0=key,
            values=t,
            indices=indt,
            ind_strict=ind_strict,
            dim=dim,
        )

    def get_time_common(
        self,
        keys=None,
        t=None,
        indt=None,
        ind_strict=None,
        dim=None,
    ):
        """ Return the time vector or time macthing indices

        hastime, hasvect, t, dind = self.get_time_common(
            keys=['prof0', 'prof1'],
            t=np.linspace(0, 5, 10),
        )

        Return
        ------
        hastime:        bool
            flag, True if key has a time dimension
        keyt:           None /  str
            if hastime and a time vector exists, the key to that time vector
        t:              None / np.ndarray
            if hastime
        indt:           None / np.ndarray
            if indt or t was provided, and keyt exists
            int indices of nearest matching times
        indtu:          None / np.ndarray
            if indt is returned, np.unique(indt)
        indt_reverse:   None / np.ndarray
            if indt is returned, a bool (ntu, nt) array

        """

        if dim is None:
            dim = 'time'

        return self.get_ref_vector_common(
            keys=keys,
            values=t,
            indices=indt,
            ind_strict=ind_strict,
            dim=dim,
        )

    def plot_as_profile2d(
        self,
        key=None,
        # parameters
        dres=None,
        dunique_mesh_2d=None,
        # levels
        dlevels=None,
        ref_com=None,
        # options
        plot_details=None,
        plot_config=None,
        # ref vectors
        dref_vectorZ=None,
        dref_vectorU=None,
        ref_vector_strategy=None,
        uniform=None,
        # interpolation
        val_out=None,
        nan0=None,
        # plot options
        dvminmax=None,
        cmap=None,
        dax=None,
        dmargin=None,
        fs=None,
        dcolorbar=None,
        dleg=None,
        # interactivity
        dinc=None,
        connect=True,
    ):

        # Plot profile 2d
        dax, dgroup = super().plot_as_profile2d(
            key=key,
            # parameters
            dres=dres,
            dunique_mesh_2d=dunique_mesh_2d,
            # levels
            dlevels=dlevels,
            ref_com=ref_com,
            # details
            plot_details=plot_details,
            # ref vectors
            dref_vectorZ=dref_vectorZ,
            dref_vectorU=dref_vectorU,
            ref_vector_strategy=ref_vector_strategy,
            uniform=uniform,
            # interpolation
            val_out=val_out,
            nan0=nan0,
            # plot options
            dvminmax=dvminmax,
            cmap=cmap,
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            dcolorbar=dcolorbar,
            dleg=dleg,
            connect=False,
        )

        # plot config if relevant
        if plot_config.__class__.__name__ == 'Config':
            if 'matrix' in dax.dax.keys():
                ax = dax.dax['matrix']['handle']
                ax = plot_config.plot(lax=ax, proj='cross')

        # -----------
        # connect

        if connect is True:
            dax.setup_interactivity(kinter='inter0', dgroup=dgroup, dinc=dinc)
            dax.disconnect_old()
            dax.connect()

            dax.show_commands()
            return dax
        else:
            return dax, dgroup

    def plot_as_profile2d_compare(
        self,
        keys=None,
        # parameters
        dres=None,
        # levels
        dlevels=None,
        ref_com=None,
        # options
        plot_config=None,
        plot_details=None,
        # ref vectors
        dref_vectorZ=None,
        dref_vectorU=None,
        ref_vector_strategy=None,
        # interpolation
        val_out=None,
        nan0=None,
        # plot options
        vmin=None,
        vmax=None,
        cmap=None,
        dax=None,
        dmargin=None,
        fs=None,
        dcolorbar=None,
        dleg=None,
        # interactivity
        dinc=None,
        connect=True,
    ):

        # Plot profile 2d
        dax, dgroup = super().plot_as_profile2d_compare(
            keys=keys,
            # parameters
            dres=dres,
            # levels
            dlevels=dlevels,
            ref_com=ref_com,
            # details
            plot_details=plot_details,
            # ref vectors
            dref_vectorZ=dref_vectorZ,
            dref_vectorU=dref_vectorU,
            ref_vector_strategy=ref_vector_strategy,
            # interpolation
            val_out=val_out,
            nan0=nan0,
            # plot options
            vmin=vmin,
            vmax=vmax,
            cmap=cmap,
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            dcolorbar=dcolorbar,
            dleg=dleg,
            # interactivity
            connect=False,
        )

        # plot config if relevant
        if plot_config.__class__.__name__ == 'Config':
            for kax in ['prof0', 'prof1']:
                if kax in dax.dax.keys():
                    ax = dax.dax[kax]['handle']
                    ax = plot_config.plot(lax=ax, proj='cross')

        # -----------
        # connect

        if connect is True:
            dax.setup_interactivity(kinter='inter0', dgroup=dgroup, dinc=dinc)
            dax.disconnect_old()
            dax.connect()

            dax.show_commands()
            return dax
        else:
            return dax, dgroup


# -*- coding: utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
# Also if needed: retab
'''
    TEST equimap
'''
from __future__ import (unicode_literals, absolute_import,  \
                        print_function, division)
import matplotlib.pyplot as plt
import numpy as np
import os
import sys

if __name__ == '__main__':
    #print('path 1 =', sys.path)
    sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
    #print('path 2 =', sys.path)

    # Local modules
    import equimap
    import imas

    interp_points = 60
    shot = 53221

    approx_time_in = 39.9881
    eps_time = 1E-3
    ctr_plot = [0]

    # CALL EQUIMAP
    oute = equimap.get(shot, time=[35.9249267578125, 39.98], \
            R=[2., 2.2, 2.4, 2.7, 3.], Phi=[0, 0, 0, 0, 0], Z=[0, 0, 0, 0, 0], \
            quantity='psi')

    print('')
    print('oute.shape =', oute.shape)
    # oute should be:
    # array([[-2.27165658, -3.01387112, -3.98753036, -3.92317787, -2.43752966],
    #        [ 1.09062263,  0.34659524, -0.63331812, -0.47944498,  1.01023176]])
    print('oute =', oute)
    print('')

    idd = imas.ids(shot, 0)
    idd.open_env('imas_public', 'west', '3')
    idd.equilibrium.get()
    out = idd.equilibrium

    equiDict = {}

    # Declaration of arrays 2d plots
    equi_grid = idd.equilibrium.grids_ggd[0].grid[0]
    NbrPoints = len(equi_grid.space[0].objects_per_dimension[0].object)
    equiDict['r'] = np.full(NbrPoints, np.nan)
    equiDict['z'] = np.full(NbrPoints, np.nan)
    for ii in range(NbrPoints):
        equiDict['r'][ii] = equi_grid.space[0].objects_per_dimension[0]. \
                            object[ii].geometry[0]
        equiDict['z'][ii] = equi_grid.space[0].objects_per_dimension[0]. \
                            object[ii].geometry[1]

    arg_time = np.argmin(np.abs(out.time - approx_time_in))

    print('\nDistance between requested and calculated time =', \
          np.abs(out.time[arg_time] - approx_time_in), '\n')

    time_in = out.time[arg_time] - eps_time

    print('\nNew distance between requested and calculated time =', \
          np.abs(out.time[arg_time] - time_in), '\n', \
          'with eps_time =', eps_time, '\n')

    R_all = np.linspace(np.min(equiDict['r']), np.max(equiDict['r']), interp_points)
    Z_all = np.linspace(np.min(equiDict['z']), np.max(equiDict['z']), interp_points)

    R_all_tot = np.repeat(R_all, interp_points)
    Z_all_tot = np.tile(Z_all, interp_points)

    # CALL EQUIMAP
    outa = equimap.get(shot, time=time_in, \
            R=R_all_tot, Phi=np.zeros(R_all_tot.shape), Z=Z_all_tot, \
            quantity='psi')

    outar = outa.reshape((interp_points, interp_points))
    Rr = R_all_tot.reshape((interp_points, interp_points))
    Zr = Z_all_tot.reshape((interp_points, interp_points))

    print('')
    print('outa.shape =', outa.shape)
    print('outar.shape =', outar.shape)
    #print('oute =', oute)
    print('')

    #plt.figure()
    #cs_ref = plt.contour(out.interp2D.r, out.interp2D.z, \
    #            out.interp2D.psi[np.argmin(np.abs(out.time - time_in))], \
    #            ctr_plot, colors='k', linestyle='solid')
    #cs_ref_tri = plt.tricontour(equiDict['r'], equiDict['z'], \
    #                 out.ggd[0].grid.space[0].objects_per_dimension[1].nodes, \
    #                 out.ggd[0].psi[np.argmin(np.abs(out.time - time_in))], \
    #                 ctr_plot, colors='red', linestyles='dashdot')
    #cs     = plt.contour(Rr, Zr, outar, ctr_plot, \
    #                     colors='orange', linestyles='dashed')
    #plt.plot(np.nan, np.nan, color='k', linestyle='solid', label='ref interp2D')
    #plt.plot(np.nan, np.nan, color='red', linestyle='dashdot', label='ref ggd')
    #plt.plot(np.nan, np.nan, color='orange', linestyle='dashed', label='test interp')
    #plt.legend()
    #plt.xlabel('R [m]')
    #plt.ylabel('Z [m]')
    #plt.title('Contour psi, level ' + str(ctr_plot) + ' at time ' + str(time_in) + 's')
    #plt.axes().set(aspect=1)
    #
    #p = cs.collections[0].get_paths()[0]
    #v = p.vertices
    #x_cs = v[:, 0]
    #y_cs = v[:, 1]

    #p = cs_ref.collections[0].get_paths()[0]
    #v = p.vertices
    #x_cs_ref = v[:, 0]
    #y_cs_ref = v[:, 1]

    plt.figure()
    plt.contourf(Rr, Zr, outar)
    plt.plot(np.squeeze(out.time_slice[arg_time].boundary.outline.r), \
             np.squeeze(out.time_slice[arg_time].boundary.outline.z), \
             linewidth=2, color='red')
    plt.plot(out.time_slice[arg_time].global_quantities.magnetic_axis.r, \
             out.time_slice[arg_time].global_quantities.magnetic_axis.z, \
             marker='+', color='red', markersize=20)
    plt.colorbar()
    plt.xlabel('R [m]')
    plt.ylabel('Z [m]')
    plt.title('psi interpolated at time ' + str(time_in) + 's')
    plt.axes().set(aspect=1)

    #plt.figure()
    #plt.plot(x_cs_ref, y_cs_ref, label='Ref')
    #plt.plot(x_cs, y_cs, '--', label='Interp')
    #plt.xlabel('R [m]')
    #plt.ylabel('Z [m]')
    #plt.legend()

    #pts_cs_ref = np.linspace(0, 1, x_cs_ref.shape[0])
    #pts_cs     = np.linspace(0, 1, x_cs.shape[0])

    #x_cs_intp = np.interp(pts_cs_ref, pts_cs, x_cs)
    #y_cs_intp = np.interp(pts_cs_ref, pts_cs, y_cs)

    #error_x = np.abs((x_cs_intp - x_cs_ref)/x_cs_ref)*100
    #error_y = np.abs((y_cs_intp - y_cs_ref)/y_cs_ref)*100

    #plt.figure()
    #plt.plot(error_x, label='err. x')
    #plt.plot(error_y, label='err. y')
    #plt.legend()
    #plt.ylabel('Error interp psi contour level ' + str(ctr_plot) + ' in [%]' \
    #          + '\nwith time difference, eps_time = ' + str(eps_time*100) + '%', \
    #          multialignment='center')

    # CALL EQUIMAP
    out_rho_pol_norm = equimap.get(shot, time=time_in, \
            R=R_all_tot, Phi=np.zeros(R_all_tot.shape), Z=Z_all_tot, \
            quantity='rho_pol_norm')

    print('')
    print('out_rho_pol_norm.shape =', out_rho_pol_norm.shape)
    #print('out_rho_pol_norm =', out_rho_pol_norm)
    print('')

    plt.figure()
    out_rho_pol_norm_r = out_rho_pol_norm.reshape((interp_points, interp_points))
    cs = plt.contourf(Rr, Zr, out_rho_pol_norm_r)
    plt.colorbar()
    plt.plot(np.squeeze(out.time_slice[arg_time].boundary.outline.r), \
             np.squeeze(out.time_slice[arg_time].boundary.outline.z), \
             linewidth=2, color='red')
    plt.plot(out.time_slice[arg_time].global_quantities.magnetic_axis.r, \
             out.time_slice[arg_time].global_quantities.magnetic_axis.z, \
             marker='+', color='red', markersize=20)
    plt.xlabel('R [m]')
    plt.ylabel('Z [m]')
    plt.title('rho_pol_norm')

    # CALL EQUIMAP
    out_rho_tor_norm = equimap.get(shot, time=time_in, \
            R=R_all_tot, Phi=np.zeros(R_all_tot.shape), Z=Z_all_tot, \
            quantity='rho_tor_norm')
    print('')
    print('out_rho_tor_norm.shape =', out_rho_tor_norm.shape)
    #print('out_rho_tor_norm =', out_rho_tor_norm)
    print('')
    plt.figure()
    out_rho_tor_norm_r = out_rho_tor_norm.reshape((interp_points, interp_points))
    cs = plt.contourf(Rr, Zr, out_rho_tor_norm_r)
    plt.colorbar()
    arg_time = np.argmin(np.abs(out.time - time_in))
    plt.plot(np.squeeze(out.time_slice[arg_time].boundary.outline.r), \
             np.squeeze(out.time_slice[arg_time].boundary.outline.z), \
             linewidth=2, color='red')
    plt.plot(out.time_slice[arg_time].global_quantities.magnetic_axis.r, \
             out.time_slice[arg_time].global_quantities.magnetic_axis.z, \
             marker='+', color='red', markersize=20)
    plt.xlabel('R [m]')
    plt.ylabel('Z [m]')
    plt.title('rho_tor_norm')

    # CALL EQUIMAP
    out_rho_tor = equimap.get(shot, time=time_in, \
            R=R_all_tot, Phi=np.zeros(R_all_tot.shape), Z=Z_all_tot, \
            quantity='rho_tor')

    print('')
    print('out_rho_tor.shape =', out_rho_tor.shape)
    #print('out_rho_tor =', out_rho_tor)
    print('')
    plt.figure()
    out_rho_tor_r = out_rho_tor.reshape((interp_points, interp_points))
    cs = plt.contourf(Rr, Zr, out_rho_tor_r)
    plt.colorbar()
    arg_time = np.argmin(np.abs(out.time - time_in))
    plt.plot(np.squeeze(out.time_slice[arg_time].boundary.outline.r), \
             np.squeeze(out.time_slice[arg_time].boundary.outline.z), \
             linewidth=2, color='red')
    plt.plot(out.time_slice[arg_time].global_quantities.magnetic_axis.r, \
             out.time_slice[arg_time].global_quantities.magnetic_axis.z, \
             marker='+', color='red', markersize=20)
    plt.xlabel('R [m]')
    plt.ylabel('Z [m]')
    plt.title('rho_tor')

    plt.figure()
    plt.plot(out.time_slice[arg_time].profiles_1d.psi, out.time_slice[arg_time].profiles_1d.rho_tor)
    plt.xlabel('Psi')
    plt.ylabel('rho_tor')

    #plt.figure()
    #plt.tricontourf(equiDict['r'], equiDict['z'], \
    #                out.ggd[0].grid.space[0].objects_per_dimension[1].nodes, \
    #                out.ggd[0].psi[arg_time])
    #plt.plot(np.squeeze(out.boundary.outline.r[arg_time]), \
    #         np.squeeze(out.boundary.outline.z[arg_time]), \
    #         linewidth=2, color='red')
    #plt.plot(out.global_quantities.magnetic_axis.r[arg_time], \
    #         out.global_quantities.magnetic_axis.z[arg_time], \
    #         marker='+', color='red', markersize=20)
    #plt.colorbar()
    #plt.xlabel('R [m]')
    #plt.ylabel('Z [m]')
    #plt.title('psi ggd')

    # CALL EQUIMAP
    out_b_norm = equimap.get(shot, time=time_in, \
            R=R_all_tot, Phi=np.zeros(R_all_tot.shape), Z=Z_all_tot, \
            quantity='b_field_norm')

    print('')
    print('out_b_norm.shape =', out_b_norm.shape)
    #print('out_b_norm =', out_b_norm)
    print('')
    plt.figure()
    out_b_norm_r = out_b_norm.reshape((interp_points, interp_points))
    cs = plt.contourf(Rr, Zr, out_b_norm_r)
    plt.colorbar()
    arg_time = np.argmin(np.abs(out.time - time_in))
    plt.plot(np.squeeze(out.time_slice[arg_time].boundary.outline.r), \
             np.squeeze(out.time_slice[arg_time].boundary.outline.z), \
             linewidth=2, color='red')
    plt.plot(out.time_slice[arg_time].global_quantities.magnetic_axis.r, \
             out.time_slice[arg_time].global_quantities.magnetic_axis.z, \
             marker='+', color='red', markersize=20)
    plt.xlabel('R [m]')
    plt.ylabel('Z [m]')
    plt.title('b_norm')

    # CALL EQUIMAP
    out_theta = equimap.get(shot, time=time_in, \
            R=R_all_tot, Phi=np.zeros(R_all_tot.shape), Z=Z_all_tot, \
            quantity='theta')

    print('')
    print('out_theta.shape =', out_theta.shape)
    print('out_theta =', out_theta)
    print('')
    plt.figure()
    out_theta_r = out_theta.reshape((interp_points, interp_points))
    cs = plt.contourf(Rr, Zr, out_theta_r, 100)
    plt.colorbar()
    arg_time = np.argmin(np.abs(out.time - time_in))
    plt.plot(np.squeeze(out.time_slice[arg_time].boundary.outline.r), \
             np.squeeze(out.time_slice[arg_time].boundary.outline.z), \
             linewidth=2, color='red')
    plt.plot(out.time_slice[arg_time].global_quantities.magnetic_axis.r, \
             out.time_slice[arg_time].global_quantities.magnetic_axis.z, \
             marker='+', color='red', markersize=20)
    plt.xlabel('R [m]')
    plt.ylabel('Z [m]')
    plt.title('theta')

    plt.show()




# class definitions (object-oriented codes)


# Built-ins
import os
import warnings

# Standard
import numpy as np
import scipy as scp
import matplotlib.pyplot as plt

# tofu-specific
import tf.__version__ as __version__

# dumpro-specific
import _comp

# !!!! never use 'from ... import *'
# Always remain explicit !!! (e.g.: np.cos() and not cos())


##########################################################
##########################################################
#               Title
##########################################################





# tofu-specific
import tofu.physics_tools as tfpt


#######################################################
#
#     Setup and Teardown
#
#######################################################


def setup_module():
    pass


def teardown_module():
    pass


#######################################################
#
#     checking Distributions
#
#######################################################


class Test00_Geometry():

    @classmethod
    def setup_class(cls):
        pass

    def setup_method(self):
        pass

    def teardown_method(self):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def test01_pinhole_camera_estimator(self):
        for col, div in [(False, False), (True, False), (True, True)]:
            dout = tfpt.geometry.camera1d_estimator(
                collimator=col,
                diverging=div,
            )
            assert isinstance(dout, dict)
